from typing import Optional

import faiss
import torch
from torch import Tensor, nn

from detectors.methods.mahalanobis import mahalanobis_distance_inv
from detectors.methods.templates import DetectorWithFeatureExtraction

HYPERPARAMETERS = dict(nclusters=dict(low=1, high=2048, step=1))


class SSD(DetectorWithFeatureExtraction):
    """SSD: A Unified Framework for Self-Supervised Outlier Detection.

    Extract features from the last layer of a self supervised model.

    References:
        [1] https://arxiv.org/abs/2103.12051
        [2] https://github.com/inspire-group/SSD
    """

    def __init__(
        self, model: nn.Module, nclusters: Optional[int] = None, niter: int = 300, cov_reg=1e-8, **kwargs
    ) -> None:
        super().__init__(
            model,
            features_nodes=[list(model._modules.keys())[-1]],
            all_blocks=False,
            last_layer=False,
            pooling_op_name="none",
            aggregation_method_name="none",
        )
        self.nclusters = nclusters
        self.niter = niter
        self.cov_reg = cov_reg

        self.kmeans = {}

    @torch.no_grad()
    def _fit_params(self) -> None:
        if self.nclusters is None:
            self.nclusters = torch.unique(self.train_targets).shape[0]

        self.kmeans = {}
        self.precisions = {}
        self.mus = {}
        self.global_mu = {}
        self.global_std = {}
        for k, v in self.train_features.items():
            # pre-processing
            v = v / (torch.norm(v, dim=-1, keepdim=True) + 1e-10)
            self.global_mu[k] = torch.mean(v, dim=0, keepdim=True)
            self.global_std[k] = torch.std(v, dim=0, keepdim=True)
            v = (v - self.global_mu[k]) / (self.global_std[k] + 1e-10)

            self.kmeans[k] = faiss.Kmeans(v.shape[1], self.nclusters, niter=self.niter, verbose=False, gpu=False)
            self.kmeans[k].train(v.cpu().numpy())

            # compute covariances
            _, ypred = self.kmeans[k].assign(v.cpu().numpy())
            self.precisions[k] = []
            self.mus[k] = []
            for i in range(self.nclusters):
                cov = torch.cov(v[ypred == i].T)
                self.precisions[k].append(torch.pinverse(cov + self.cov_reg * torch.eye(cov.shape[1], device=v.device)))
                self.mus[k].append(torch.mean(v[ypred == i], dim=0))

        del self.kmeans

    @torch.no_grad()
    def _layer_score(self, features: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None, **kwargs):
        features = features / (torch.norm(features, dim=-1, keepdim=True) + 1e-10)
        features = (features - self.global_mu[layer_name].to(features.device)) / (
            self.global_std[layer_name].to(features.device) + 1e-10
        )
        scores = torch.zeros((features.shape[0], len(self.mus[layer_name])), device=features.device)
        for i, (mu, pinv) in enumerate(zip(self.mus[layer_name], self.precisions[layer_name])):
            scores[:, i] = mahalanobis_distance_inv(features, mu.to(features.device), pinv.to(features.device))
        return -torch.nan_to_num(scores.min(dim=1)[0], 1e6)



import os
import subprocess
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, extract_archive


class TinyImageNetResized(ImageFolder):
    base_folder = "Imagenet_resized"
    filename = "Imagenet_resize.tar.gz"
    file_md5 = "0f9ff11d45babf2eff5fe12281d1ac31"
    url = "https://www.dropbox.com/s/kp3my3412u5k9rl/Imagenet_resize.tar.gz"

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)
        self.dataset_folder = os.path.join(self.root, self.base_folder)
        self.archive = os.path.join(self.root, self.filename)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        super().__init__(self.dataset_folder, transform=transform, **kwargs)

    def _check_integrity(self) -> bool:
        return check_integrity(self.archive, self.file_md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self.dataset_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        print(f"Downloading {self.filename}...")
        subprocess.run(f"wget {self.url} -P {self.root}".split(" "), capture_output=True, text=True)
        extract_archive(self.archive, self.dataset_folder)


class TinyImageNetCroped(TinyImageNetResized):
    base_folder = "Imagenet_croped"
    filename = "Imagenet.tar.gz"
    file_md5 = "7c0827e4246c3718a5ee076e999e52e5"
    url = "https://www.dropbox.com/s/avgm2u562itwpkl/Imagenet.tar.gz"



import logging

import torch
from torch import Tensor

_logger = logging.getLogger(__name__)


class InnerProductAggregation:
    def __init__(self, *args, **kwargs) -> None:
        self.max_trajectory = None
        self.ref_trajectory = None
        self.scale = None

    def fit(self, stack: Tensor, *args, **kwargs):
        self.max_trajectory = stack.max(dim=0, keepdim=True)[0]
        self.ref_trajectory = stack.mean(dim=0, keepdim=True) / self.max_trajectory
        self.scale = torch.sum(self.ref_trajectory**2)

        _logger.debug("InnerProductAggregation parameters")
        _logger.debug(f"max_trajectory: {self.max_trajectory}")
        _logger.debug(f"ref_trajectory: {self.ref_trajectory}")
        _logger.debug(f"scale: {self.scale}")

    def __call__(self, scores: Tensor, *args, **kwargs):
        scores = scores / self.max_trajectory.to(scores.device)
        scores = torch.sum(scores * self.ref_trajectory.to(scores.device), dim=1) / self.scale.to(scores.device)
        return scores



import logging
from typing import Optional

import torch
from torch import Tensor

from .mahalanobis import Mahalanobis, mahalanobis_distance_inv

_logger = logging.getLogger(__name__)


def relative_mahalanobis_inv_layer_score(
    x: Tensor, mus: Tensor, inv: Tensor, background_mu: Tensor, background_inv: Tensor
):
    stack = torch.zeros((x.shape[0], mus.shape[0]), device=x.device, dtype=torch.float32)
    for i, mu in enumerate(mus):
        stack[:, i] = mahalanobis_distance_inv(x, mu.reshape(1, -1), inv).reshape(-1) - mahalanobis_distance_inv(
            x, background_mu.reshape(1, -1), background_inv
        ).reshape(-1)

    return -torch.nan_to_num(stack.min(1, keepdim=True)[0], nan=1e6)


HYPERPARAMETERS = dict(
    cov_mat_method=[
        "EmpiricalCovariance",
        "GraphicalLasso",
        "GraphicalLassoCV",
        "LedoitWolf",
        "ShrunkCovariance",
        "OAS",
    ]
)


class RelativeMahalanobis(Mahalanobis):
    """RelativeMahalanobis detector.

    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of ["max", "avg", "flatten", "getitem", "none"]. Defaults to "avg".
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.
        cov_mat_method (str, optional): Covariance matrix estimation method. Can be one of
            ["EmpiricalCovariance", "GraphicalLasso", "GraphicalLassoCV", "LedoitWolf", "MinCovDet", "ShrunkCovariance", "OAS"].
            Defaults to "EmpiricalCovariance".
        mu_cov_inv_est_fn (function, optional): Function to be used to estimate the means, covariance and inverse matrix.
            Defaults to `class_cond_mus_cov_inv_matrix`.
        cov_reg (float, optional): Covariance regularization. Defaults to 1e-6.

    References:
        [1] https://arxiv.org/abs/2106.03004
    """

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        return relative_mahalanobis_inv_layer_score(
            x,
            self.mus[layer_name].to(x.device),
            self.invs[layer_name].to(x.device),
            self.background_mus[layer_name].to(x.device),
            self.background_invs[layer_name].to(x.device),
        )

    def _fit_params(self) -> None:
        self.mus = {}
        self.invs = {}
        self.background_mus = {}
        self.background_invs = {}
        device = next(self.model.parameters()).device
        for layer_name, layer_features in self.train_features.items():
            self.mus[layer_name], cov, self.invs[layer_name] = self.mu_cov_inv_est_fn(
                layer_features, self.train_targets, self.cov_mat_method, device=device
            )
            self.background_mus[layer_name] = torch.mean(layer_features, dim=0, keepdim=True).to(device)
            self.background_invs[layer_name] = torch.pinverse(torch.cov(layer_features.T).to(device))



import os
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive

MOS_PAPER_URL = ""


class MOSSUN(ImageFolder):
    """`MOS SUN <MOS_PAPER_URL>`_ Dataset subset.

    Args:
        root (string): Root directory of dataset where directory
            exists or will be saved to if download is set to True.
        split (string, optional): The dataset split, not used.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments passed to :class:`~torchvision.datasets.ImageFolder`.
    """

    base_folder = "mos_sun"
    filename = "SUN.tar.gz"
    file_md5 = "8469c3ada62211477954ec1be53b12d0"
    url = "http://pages.cs.wisc.edu/~huangrui/imagenet_ood_dataset/SUN.tar.gz"
    # size: 10000

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)
        self.dataset_folder = os.path.join(self.root, self.base_folder)
        self.archive = os.path.join(self.root, self.filename)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        super().__init__(self.dataset_folder, transform=transform, **kwargs)

    def _check_integrity(self) -> bool:
        return check_integrity(self.archive, self.file_md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self.dataset_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        download_and_extract_archive(
            self.url, download_root=self.root, extract_root=self.dataset_folder, md5=self.file_md5
        )


class MOSPlaces365(MOSSUN):
    """`MOS Places365 <MOS_PAPER_URL>`_ Dataset subset.

    Args:
        root (string): Root directory of dataset where directory
            exists or will be saved to if download is set to True.
        split (string, optional): The dataset split, not used.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments passed to :class:`~torchvision.datasets.ImageFolder`.
    """

    base_folder = "mos_places365"
    filename = "Places.tar.gz"
    file_md5 = "b5cb5eba2754ae2a28beea8718db699a"
    url = "http://pages.cs.wisc.edu/~huangrui/imagenet_ood_dataset/Places.tar.gz"


class MOSiNaturalist(MOSSUN):
    """`MOS iNaturalist <MOS_PAPER_URL>`_ Dataset subset.

    Args:
        root (string): Root directory of dataset where directory
            exists or will be saved to if download is set to True.
        split (string, optional): The dataset split, not used.
        transform (callable, optional): A function/transform that takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`.
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments passed to :class:`~torchvision.datasets.ImageFolder`.
    """

    base_folder = "mos_inaturalist"
    filename = "iNaturalist.tar.gz"
    file_md5 = "5be6ea8aa027d7b631916427b32cb335"
    url = "http://pages.cs.wisc.edu/~huangrui/imagenet_ood_dataset/iNaturalist.tar.gz"



"""Trainer for classification models."""
import json
import logging
import os
from typing import Callable, Tuple

import accelerate
import torch
import torch.utils.data
from accelerate import Accelerator
from torch import Tensor, nn, optim
from tqdm import tqdm

_logger = logging.getLogger(__name__)


def get_criterion_cls(criterion_name: str) -> nn.modules.loss._Loss:
    return getattr(nn, criterion_name)


def get_optimizer_cls(optimizer_name: str) -> optim.Optimizer:
    return getattr(optim, optimizer_name)


def get_scheduler_cls(scheduler_name: str) -> optim.lr_scheduler._LRScheduler:
    return getattr(optim.lr_scheduler, scheduler_name)


def training_iteration(
    batch: Tuple[Tensor, Tensor],
    model: nn.Module,
    optimizer: optim.Optimizer,
    criterion: Callable,
    accelerator: Accelerator,
):
    inputs, targets = batch

    optimizer.zero_grad()

    outputs = model(inputs)
    loss = criterion(outputs, targets)

    accelerator.backward(loss)
    optimizer.step()
    return {"loss": loss.item()}


def validation_iteration(batch: Tuple[Tensor, Tensor], model: nn.Module, criterion: Callable, accelerator: Accelerator):
    inputs, targets = batch
    with torch.no_grad():
        outputs = model(inputs)
    predictions = torch.argmax(outputs, dim=1)
    predictions, targets = accelerator.gather_for_metrics((predictions, targets))
    loss = criterion(outputs, targets).item()
    acc = (predictions == targets).float().sum().item()
    return {"loss": loss, "accuracy": acc}


def save_model(model: nn.Module, accelerator: Accelerator, filename: str):
    accelerator.wait_for_everyone()
    unwrapped_model = accelerator.unwrap_model(model)
    accelerator.save(unwrapped_model.state_dict(), filename)


def trainer_supervised_classification(
    model: torch.nn.Module,
    optimizer: torch.optim.Optimizer,
    scheduler: torch.optim.lr_scheduler._LRScheduler,
    criterion: Callable,
    train_loader: torch.utils.data.DataLoader,
    val_loader: torch.utils.data.DataLoader,
    save_root: str,
    training_function=training_iteration,
    validation_function=validation_iteration,
    epochs=10,
    validation_frequency=1,
    seed: int = 42,
):
    os.makedirs(save_root, exist_ok=True)
    _logger.info(f"Saving model and progress to {save_root}")

    accelerate.utils.set_seed(seed)
    accelerator = Accelerator(
        log_with=["all"], logging_dir=os.path.join(save_root, "logs"), step_scheduler_with_optimizer=False
    )
    accelerator.init_trackers("")
    model, optimizer, train_loader, val_loader, scheduler = accelerator.prepare(
        model, optimizer, train_loader, val_loader, scheduler
    )

    best_accuracy = 0.0
    step = 0
    progress_bar = tqdm(range(epochs), disable=not accelerator.is_local_main_process, dynamic_ncols=True)
    for epoch in progress_bar:
        # train
        model.train()
        for batch in train_loader:
            step += 1
            tr_obj = training_function(batch, model, optimizer, criterion, accelerator)
            tr_loss = tr_obj["loss"]
            lr = optimizer.param_groups[0]["lr"]
            progress_bar.set_description_str(f"{step}it, loss={tr_loss:.4f}, lr={lr:.4f}")
            accelerator.log({f"train/{k}": v for k, v in tr_obj.items()}, step=step)
            accelerator.log({"lr": lr}, step=step)

        # validate
        if (epoch + 1) % validation_frequency == 0 or epoch == 0 or epoch == epochs - 1:
            model.eval()
            val_acc = 0
            val_loss = 0
            total = 0
            val_obj = {}
            for batch in val_loader:
                val_obj = validation_function(batch, model, criterion, accelerator)

                val_acc += val_obj["accuracy"]
                val_loss += val_obj["loss"]
                total += len(batch[0])  # temporary fix

            val_acc /= total
            val_loss /= total

            if val_acc > best_accuracy:
                best_accuracy = val_acc
                save_model(model, accelerator, os.path.join(save_root, "best.pth"))

            progress_bar.set_postfix({"val/loss": val_loss, "val/acc": val_acc, "best/acc": best_accuracy})
            accelerator.log({"val/acc": val_acc, "val/loss": val_loss}, step=epoch)

            if scheduler is not None:
                if isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                    scheduler.step(val_acc)

            # finish training if model is not learning
            num_classes = model(batch[0]).shape[1]  # temporary fix
            if val_acc < 1 / num_classes:
                _logger.error("Accuracy is too low, stopping training")
                break

        # sync accelerator after epoch
        accelerator.wait_for_everyone()
        progress_bar.update(1)

        if scheduler is not None:
            if not isinstance(scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
                scheduler.step(epoch)

    accelerator.end_training()
    # save model
    save_model(model, accelerator, os.path.join(save_root, "last.pth"))

    # save train results
    train_results = {"epoch": epoch, "train_loss": tr_loss}
    with open(os.path.join(save_root, "train_results.json"), "w") as f:
        json.dump(train_results, f)

    # save eval results
    eval_results = {"epoch": epoch, "best_accuracy": best_accuracy, "last_accuracy": val_acc, "eval_loss": val_loss}
    with open(os.path.join(save_root, "eval_results.json"), "w") as f:
        json.dump(eval_results, f)

    _logger.info(f"Training finished. Best accuracy: {best_accuracy:.4f}")
    _logger.info(f"Last model saved to {save_root}/last.pth")
    _logger.info(f"Best model saved to {save_root}/best.pth")
    _logger.info(f"Training logs saved to {save_root}/logs")
    _logger.info(f"Training results saved to {save_root}/train_results.json")
    _logger.info(f"Evaluation results saved to {save_root}/eval_results.json")



import torch
from torch import Tensor


def none_aggregation(x: Tensor, *args, **kwargs) -> Tensor:
    return x


def mean_aggregation(x: Tensor, *args, **kwargs) -> Tensor:
    return x.mean(dim=1, keepdim=True)


def median_aggregation(x: Tensor, *args, **kwargs) -> Tensor:
    return x.median(dim=1, keepdim=True)[0]


def max_aggregation(x: Tensor, *args, **kwargs) -> Tensor:
    return x.max(dim=1, keepdim=True)[0]


def min_aggregation(x: Tensor, *args, **kwargs) -> Tensor:
    return x.min(dim=1, keepdim=True)[0]


def avg_topk_aggregation(x: Tensor, k: int, *args, **kwargs) -> Tensor:
    return x.topk(k, dim=1)[0].mean(dim=1, keepdim=True)


def topk_aggregation(x: Tensor, k: int, *args, **kwargs) -> Tensor:
    return x.topk(k, dim=1)[0]


def layer_idx(x: Tensor, layer_idx: int = -1, *args, **kwargs) -> Tensor:
    return x[:, layer_idx]


def depth_weighted_sum(x: Tensor, *args, **kwargs) -> Tensor:
    w = torch.arange(0, x.shape[1], 1, device=x.device, dtype=x.dtype) / x.shape[1]
    x = x * w
    return x.sum(dim=1, keepdim=True)



"""
OOD Pipelines.
"""
import logging
import time
from abc import ABC, abstractmethod
from typing import Any, Callable, Dict, List, Literal, Tuple, Union

import accelerate
import numpy as np
import optuna
import pandas as pd
import torch
import torch.utils.data
import torchvision
from optuna.trial import TrialState
from torch import Tensor
from tqdm import tqdm

from detectors.data import create_dataset
from detectors.eval import get_ood_results
from detectors.methods import DetectorWrapper
from detectors.pipelines import register_pipeline
from detectors.pipelines.base import Pipeline
from detectors.utils import ConcatDatasetsDim1, sync_tensor_across_gpus

_logger = logging.getLogger(__name__)


METRICS_NAMES_PRETTY = {
    "fpr_at_0.95_tpr": "FPR at 95% TPR",
    "tnr_at_0.95_tpr": "TNR at 95% TPR",
    "detection_error": "Detection error",
    "auroc": "AUROC",
    "aupr_in": "AUPR in",
    "aupr_out": "AUPR out",
    "thr": "Threshold",
    "time": "Time",
}


class OODBenchmarkPipeline(Pipeline, ABC):
    """OOD Benchmark pipeline.

    Args:
        in_dataset_name (str): Name of the in-distribution dataset.
        out_datasets_names_splits (Dict[str, Any]): Dictionary mapping out-distribution dataset names to their splits.
        transform (Callable): Transform to apply to the datasets.
        batch_size (int): Batch size.
        num_workers (int, optional): Number of workers. Defaults to 4.
        pin_memory (bool, optional): Pin memory. Defaults to True.
        prefetch_factor (int, optional): Prefetch factor. Defaults to 2.
        limit_fit (float, optional): Fraction of the training set to use for fitting. Defaults to 1.0.
        limit_run (float, optional): Fraction of the testing set to use for running. Defaults to 1.0.
        seed (int, optional): Random seed. Defaults to 42.
        accelerator (Any, optional): Accelerator. Defaults to None.
    """

    def __init__(
        self,
        in_dataset_name: str,
        out_datasets_names_splits: Dict[str, Any],
        transform: Callable,
        batch_size: int,
        num_workers: int = 4,
        pin_memory: bool = True,
        prefetch_factor: int = 2,
        limit_fit: float = 1.0,
        limit_run: float = 1.0,
        seed: int = 42,
        accelerator=None,
    ) -> None:
        self.in_dataset_name = in_dataset_name
        self.out_datasets_names_splits = out_datasets_names_splits
        self.out_datasets_names = list(out_datasets_names_splits.keys())
        self.limit_fit = limit_fit
        self.limit_run = limit_run
        self.transform = transform
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.prefetch_factor = prefetch_factor
        self.seed = seed
        self.accelerator = accelerator

        self.fit_dataset = None
        self.in_dataset = None
        self.out_dataset = None
        self.out_datasets = None

        accelerate.utils.set_seed(seed)
        print("Setting up datasets...")
        self.setup()

    @abstractmethod
    def _setup_datasets(self):
        """Setup `in_dataset`, `out_dataset`, `fit_dataset` and `out_datasets`."""
        ...

    def _setup_dataloaders(self):
        if self.fit_dataset is None or self.in_dataset is None or self.out_datasets is None or self.out_dataset is None:
            raise ValueError("Datasets are not set.")

        if self.limit_fit is None:
            self.limit_fit = 1.0
        self.limit_fit = min(int(self.limit_fit * len(self.fit_dataset)), len(self.fit_dataset))

        # random indices
        subset = np.random.choice(np.arange(len(self.fit_dataset)), self.limit_fit, replace=False).tolist()
        self.fit_dataset = torch.utils.data.Subset(self.fit_dataset, subset)
        self.fit_dataloader = torch.utils.data.DataLoader(
            self.fit_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            prefetch_factor=self.prefetch_factor,
        )

        self.test_dataset = torch.utils.data.ConcatDataset([self.in_dataset, self.out_dataset])
        test_labels = torch.utils.data.TensorDataset(
            torch.cat(
                [torch.zeros(len(self.in_dataset))]  # type: ignore
                + [torch.ones(len(d)) * (i + 1) for i, d in enumerate(self.out_datasets.values())]  # type: ignore
            ).long()
        )

        self.test_dataset = ConcatDatasetsDim1([self.test_dataset, test_labels])
        # shuffle and subsample test_dataset
        subset = np.random.choice(
            np.arange(len(self.test_dataset)), int(self.limit_run * len(self.test_dataset)), replace=False
        ).tolist()
        self.test_dataset = torch.utils.data.Subset(self.test_dataset, subset)
        self.test_dataloader = torch.utils.data.DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            prefetch_factor=self.prefetch_factor,
        )

        if self.accelerator is not None:
            self.fit_dataloader = self.accelerator.prepare(self.fit_dataloader)
            self.test_dataloader = self.accelerator.prepare(self.test_dataloader)

        _logger.info(f"Using {len(self.fit_dataset)} samples for fitting.")
        _logger.info(f"Using {len(self.test_dataset)} samples for testing.")

    def setup(self):
        self._setup_datasets()
        self._setup_dataloaders()

    def preprocess(self, method: DetectorWrapper) -> DetectorWrapper:
        if self.fit_dataset is None:
            _logger.warning("Fit dataset is not set or not supported. Returning.")
            return method

        if not hasattr(method.detector, "update"):
            _logger.warning("Detector does not support fitting. Returning.")
            return method

        disable = False
        if self.accelerator is not None:
            disable = not self.accelerator.is_main_process
        progress_bar = tqdm(range(len(self.fit_dataloader)), desc="Fitting", disable=disable)

        fit_length = len(self.fit_dataloader.dataset)
        example = next(iter(self.fit_dataloader))[0]
        method.start(example=example, fit_length=fit_length)
        for x, y in self.fit_dataloader:
            method.update(x, y)
            progress_bar.update(1)
        progress_bar.close()
        method.end()
        return method

    def run(self, method: DetectorWrapper) -> Dict[str, Any]:
        self.method = method

        _logger.info("Running pipeline...")
        self.method = self.preprocess(self.method)

        # initialize based on dataset size
        dataset_size = len(self.test_dataloader.dataset)
        test_labels = torch.empty(dataset_size, dtype=torch.int64)
        test_scores = torch.empty(dataset_size, dtype=torch.float32)
        _logger.debug("test_labels shape: %s", test_labels.shape)
        _logger.debug("test_scores shape: %s", test_scores.shape)

        self.infer_times = []
        idx = 0
        disable = False
        if self.accelerator is not None:
            disable = not self.accelerator.is_main_process
        progress_bar = tqdm(range(len(self.test_dataloader)), desc="Inference", disable=disable)
        for x, y, labels in self.test_dataloader:
            t1 = time.time()
            score = self.method(x)
            t2 = time.time()

            if self.accelerator is not None:
                score = self.accelerator.gather_for_metrics(score)
                labels = self.accelerator.gather_for_metrics(labels)
            # score = sync_tensor_across_gpus(score.detach())
            # labels = sync_tensor_across_gpus(labels.to(score.device))

            self.infer_times.append(t2 - t1)
            test_labels[idx : idx + labels.shape[0]] = labels.cpu()
            test_scores[idx : idx + score.shape[0]] = score.cpu()

            idx += labels.shape[0]
            progress_bar.update(1)
        progress_bar.close()
        self.infer_times = np.mean(self.infer_times)
        test_scores = test_scores[:idx]
        test_labels = test_labels[:idx]
        res_obj = self.postprocess(test_scores, test_labels)

        return {"results": res_obj, "scores": test_scores, "labels": test_labels}

    def postprocess(self, test_scores: Tensor, test_labels: Tensor):
        _logger.info("Computing metrics...")
        in_scores = test_scores[test_labels == 0]

        results = {}
        for i, ood_dataset_name in enumerate(self.out_datasets_names):
            ood_scores = test_scores[test_labels == (i + 1)]
            results[ood_dataset_name] = get_ood_results(in_scores, ood_scores)
            results[ood_dataset_name]["time"] = self.infer_times

        results["average"] = {
            k: np.mean([results[ds][k] for ds in self.out_datasets_names])
            for k in results[self.out_datasets_names[0]].keys()
        }
        results["average"]["time"] = self.infer_times
        ood_scores = test_scores[test_labels > 0]

        return results

    def report(self, results: Dict[str, Dict[str, Any]]) -> str:
        # log results in a table
        if "results" in results:
            results = results["results"]
        df = pd.DataFrame()

        for ood_dataset, res in results.items():
            df = pd.concat([df, pd.DataFrame(res, index=[ood_dataset])])
        df.columns = [METRICS_NAMES_PRETTY[k] for k in df.columns]
        return df.to_string(index=True, float_format="{:.4f}".format)


@register_pipeline("ood_benchmark_cifar10")
class OODCifar10BenchmarkPipeline(OODBenchmarkPipeline):
    def __init__(self, transform: Callable, limit_fit=1.0, limit_run=1.0, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            "cifar10",
            {
                "cifar100": "test",
                "svhn": "test",
                "isun": None,
                "lsun_c": None,
                "lsun_r": None,
                "tiny_imagenet_c": None,
                "tiny_imagenet_r": None,
                "textures": None,
                "places365": None,
                "english_chars": None,
                "uniform": None,
                "gaussian": None,
            },
            transform=transform,
            batch_size=batch_size,
            limit_fit=limit_fit,
            limit_run=limit_run,
            seed=seed,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform, download=True)
        self.in_dataset = create_dataset(self.in_dataset_name, split="test", transform=self.transform, download=True)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


@register_pipeline("ood_benchmark_cifar100")
class OODCifar100BenchmarkPipeline(OODBenchmarkPipeline):
    def __init__(self, transform: Callable, limit_fit=1.0, limit_run=1.0, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            "cifar100",
            {
                "cifar10": "test",
                "svhn": "test",
                "isun": None,
                "lsun_c": None,
                "lsun_r": None,
                "tiny_imagenet_c": None,
                "tiny_imagenet_r": None,
                "textures": None,
                "places365": None,
                "english_chars": None,
                "uniform": None,
                "gaussian": None,
            },
            transform=transform,
            batch_size=batch_size,
            limit_fit=limit_fit,
            limit_run=limit_run,
            seed=seed,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform, download=True)
        self.in_dataset = create_dataset(self.in_dataset_name, split="test", transform=self.transform, download=True)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


@register_pipeline("ood_benchmark_imagenet")
class OODImageNetBenchmarkPipeline(OODBenchmarkPipeline):
    def __init__(self, transform: Callable, limit_fit=1.0, limit_run=1.0, batch_size=64, seed=42, **kwargs) -> None:
        super().__init__(
            "ilsvrc2012",
            {
                "mos_inaturalist": None,
                "mos_sun": None,
                "mos_places365": None,
                "textures": None,
                "imagenet_o": None,
                "openimage_o": None,
                "imagenet_a": None,
                "imagenet_r": None,
                "uniform": None,
                "gaussian": None,
            },
            limit_fit=limit_fit,
            limit_run=limit_run,
            transform=transform,
            batch_size=batch_size,
            seed=seed,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform)
        self.in_dataset = create_dataset(self.in_dataset_name, split="val", transform=self.transform)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


@register_pipeline("ood_mnist_benchmark")
class OODMNISTBenchmarkPipeline(OODBenchmarkPipeline):
    def __init__(self, transform: Callable, limit_fit=1, batch_size=64) -> None:
        super().__init__(
            "mnist",
            {
                "fashion_mnist": "test",
                "svhn": "test",
                "cifar10": "test",
                "textures": None,
                "english_chars": None,
            },
            limit_fit=limit_fit,
            transform=transform,
            batch_size=batch_size,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.transform.transforms.append(torchvision.transforms.Grayscale(num_output_channels=1))
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform)
        self.in_dataset = create_dataset(self.in_dataset_name, split="test", transform=self.transform)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


class OODValidationPipeline(OODBenchmarkPipeline, ABC):
    """Pipeline for OOD validation.

    This pipeline is used to validate the performance of a model on OOD datasets.

    Args:
        method (DetectorWrapper): The OOD detection method to use.
        hyperparameters (Dict[str, Union[List[Any], Tuple[Any], Dict[str, Any]]]): The hyperparameters to use for the method.
        objective_metric (Literal["fpr_at_0.95_tpr", "auroc"], optional): The metric to optimize. Defaults to "auroc".
        n_trials (int, optional): The number of trials to run. Defaults to 20.
    """

    # TODO: include prevent refit flag.

    def run(
        self,
        method: DetectorWrapper,
        hyperparameters: Dict[str, Union[List[Any], Tuple[Any], Dict[str, Any]]],
        objective_metric: Literal["fpr_at_0.95_tpr", "auroc"] = "auroc",
        objective_dataset: str = "average",
        n_trials=20,
    ) -> Dict[str, Any]:
        self.method = method
        self.hyperparameters = hyperparameters
        self.objective_metric = objective_metric
        self.objective_dataset = objective_dataset

        direction = "maximize" if objective_metric == "auroc" else "minimize"
        sampler = None
        if all(isinstance(v, (list, tuple)) for v in hyperparameters.values()):
            sampler = optuna.samplers.GridSampler(search_space=hyperparameters)
            lengths = np.array([len(v) for v in hyperparameters.values()])
            n_trials = min(int(np.prod(lengths)), n_trials)
        study = optuna.create_study(study_name="ood-val", sampler=sampler, direction=direction)
        study.optimize(self.objective, n_trials=n_trials, show_progress_bar=True)

        self.method.set_hyperparameters(**study.best_params)
        return {
            "method": self.method,
            "study": study,
            "best_params": study.best_params,
            "best_value": study.best_trial.value,
        }

    def objective(self, trial: optuna.trial.Trial) -> float:
        # build detector from trial params
        new_params = {}
        for k in self.hyperparameters:
            if isinstance(self.hyperparameters[k], (list, tuple)):
                new_params[k] = trial.suggest_categorical(k, self.hyperparameters[k])
            elif isinstance(self.hyperparameters[k], dict):
                step = self.hyperparameters[k]["step"]
                low = self.hyperparameters[k]["low"]
                high = self.hyperparameters[k]["high"]
                param_type = type(step)
                if param_type == float:
                    new_params[k] = trial.suggest_float(k, low=low, high=high, step=step)
                elif param_type == int:
                    new_params[k] = trial.suggest_int(k, low=low, high=high, step=step)

        self.method.set_hyperparameters(**new_params)
        # print methods params
        run_obj = super().run(self.method)
        results = run_obj["results"]
        return results[self.objective_dataset][self.objective_metric]

    def report(self, results: Dict[str, Any]):
        if "study" not in results:
            raise ValueError("The results dict must contain a 'study' key.")
        study = results["study"]
        pruned_trials = study.get_trials(deepcopy=False, states=[TrialState.PRUNED])
        complete_trials = study.get_trials(deepcopy=False, states=[TrialState.COMPLETE])
        best_trial = study.best_trial
        best_value = best_trial.value

        report_str = f"""
Study statistics:
    Number of finished trials: {len(study.trials)}
    Number of pruned trials: {len(pruned_trials)}
    Number of complete trials: {len(complete_trials)}
    Best trial:
        Value: {best_value}
        Params: {best_trial.params}
"""
        return report_str


@register_pipeline("ood_validation_cifar10")
class OODCifar10ValidationPipeline(OODCifar10BenchmarkPipeline, OODValidationPipeline):
    def __init__(self, transform: Callable, limit_fit=1, limit_run=0.1, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            transform=transform, batch_size=batch_size, limit_fit=limit_fit, limit_run=limit_run, seed=seed
        )


@register_pipeline("ood_validation_noise_cifar10")
class OODCifar10NoiseValidationPipeline(OODValidationPipeline):
    def __init__(self, transform: Callable, limit_fit=1, limit_run=0.1, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            "cifar10",
            {
                "uniform": None,
                "gaussian": None,
            },
            transform=transform,
            batch_size=batch_size,
            limit_fit=limit_fit,
            limit_run=limit_run,
            seed=seed,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform)
        self.in_dataset = create_dataset(self.in_dataset_name, split="test", transform=self.transform)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


@register_pipeline("ood_validation_cifar100")
class OODCifar100ValidationPipeline(OODCifar100BenchmarkPipeline, OODValidationPipeline):
    def __init__(self, transform: Callable, limit_fit=1, limit_run=0.1, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            transform=transform, batch_size=batch_size, limit_fit=limit_fit, limit_run=limit_run, seed=seed
        )


@register_pipeline("ood_validation_noise_cifar100")
class OODCifar100NoiseValidationPipeline(OODValidationPipeline):
    def __init__(self, transform: Callable, limit_fit=1, limit_run=0.1, batch_size=128, seed=42, **kwargs) -> None:
        super().__init__(
            "cifar100",
            {
                "uniform": None,
                "gaussian": None,
            },
            transform=transform,
            batch_size=batch_size,
            limit_fit=limit_fit,
            limit_run=limit_run,
            seed=seed,
        )

    def _setup_datasets(self):
        _logger.info("Loading In-distribution dataset...")
        self.fit_dataset = create_dataset(self.in_dataset_name, split="train", transform=self.transform)
        self.in_dataset = create_dataset(self.in_dataset_name, split="test", transform=self.transform)

        _logger.info("Loading OOD datasets...")
        self.out_datasets = {
            ds: create_dataset(ds, split=split, transform=self.transform, download=True)
            for ds, split in self.out_datasets_names_splits.items()
        }
        self.out_dataset = torch.utils.data.ConcatDataset(list(self.out_datasets.values()))


@register_pipeline("ood_validation_imagenet")
class OODImageNetValidationPipeline(OODImageNetBenchmarkPipeline, OODValidationPipeline):
    def __init__(self, transform: Callable, limit_fit=1, limit_run=0.1, batch_size=64, seed=42, **kwargs) -> None:
        super().__init__(
            transform=transform, batch_size=batch_size, limit_fit=limit_fit, limit_run=limit_run, seed=seed
        )



"""
Datasets module.
"""
import logging
from enum import Enum
from functools import partial
from typing import Callable, List, Optional, Type

from torch.utils.data import Dataset
from torchvision.datasets import STL10, SVHN, ImageNet, OxfordIIITPet, StanfordCars

from ..config import DATA_DIR, IMAGENET_ROOT
from .cifar_wrapper import CIFAR10Wrapped, CIFAR100Wrapped
from .cifarc import CIFAR10_C, CIFAR100_C
from .constants import *
from .english_chars import EnglishChars
from .imagenet import ImageNetA, ImageNetC, ImageNetCnpz, ImageNetO, ImageNetR
from .isun import iSUN
from .lsun_r_c import LSUNCroped, LSUNResized
from .mnist_wrapped import FashionMNISTWrapped, MNISTWrapped
from .mnistc import MNISTC
from .mos import MOSSUN, MOSiNaturalist, MOSPlaces365
from .noise import Blobs, Gaussian, Rademacher, Uniform
from .openimage_o import OpenImageO
from .places365 import Places365
from .textures import Textures
from .tiny_imagenet import TinyImageNet
from .tiny_imagenet_r_c import TinyImageNetCroped, TinyImageNetResized
from .wilds_ds import make_wilds_dataset

_logger = logging.getLogger(__name__)
datasets_registry = {
    "cifar10": CIFAR10Wrapped,
    "cifar100": CIFAR100Wrapped,
    "stl10": STL10,
    "svhn": SVHN,
    "mnist": MNISTWrapped,
    "fashion_mnist": FashionMNISTWrapped,
    "kmnist": None,
    "emnist": None,
    "mnist_c": MNISTC,
    "english_chars": EnglishChars,
    "isun": iSUN,
    "lsun_c": LSUNCroped,
    "lsun_r": LSUNResized,
    "tiny_imagenet_c": TinyImageNetCroped,
    "tiny_imagenet_r": TinyImageNetResized,
    "tiny_imagenet": TinyImageNet,
    "textures": Textures,
    "textures_curated": None,
    "gaussian": Gaussian,
    "uniform": Uniform,
    "blobs": Blobs,
    "rademacher": Rademacher,
    "places365": Places365,
    "stanford_cars": StanfordCars,
    "mos_inaturalist": MOSiNaturalist,
    "mos_places365": MOSPlaces365,
    "mos_sun": MOSSUN,
    "cifar10_lt": None,
    "cifar100_lt": None,
    "imagenet1k_lt": None,
    "cifar10_c": CIFAR10_C,
    "cifar100_c": CIFAR100_C,
    "imagenet": ImageNet,
    "imagenet1k": ImageNet,
    "ilsvrc2012": ImageNet,
    "imagenet_c": ImageNetCnpz,
    "imagenet_c_npz": ImageNetCnpz,
    "imagenet1k_c": ImageNetC,
    "imagenet_a": ImageNetA,
    "imagenet_r": ImageNetR,
    "imagenet_o": ImageNetO,
    "openimage_o": OpenImageO,
    "oxford_pets": OxfordIIITPet,
    "oxford_flowers": None,
    "cub200": None,
    "wilds_iwildcam": partial(make_wilds_dataset, dataset_name="iwildcam"),
    "wilds_fmow": partial(make_wilds_dataset, dataset_name="fmow"),
    "wilds_camelyon17": partial(make_wilds_dataset, dataset_name="camelyon17"),
    "wilds_rxrx1": partial(make_wilds_dataset, dataset_name="rxrx1"),
    "wilds_poverty": partial(make_wilds_dataset, dataset_name="poverty"),
    "wilds_globalwheat": partial(make_wilds_dataset, dataset_name="globalwheat"),
}


def register_dataset(dataset_name: str):
    """Register a dataset on the `datasets_registry`.

    Args:
        dataset_name (str): Name of the dataset.

    Example::

        @register_dataset("my_dataset")
        class MyDataset(Dataset):
            ...

        dataset = create_dataset("my_dataset")
    """

    def register_model_cls(cls):
        if dataset_name in datasets_registry:
            raise ValueError(f"Cannot register duplicate dataset ({dataset_name})")
        datasets_registry[dataset_name] = cls
        return cls

    return register_model_cls


def create_dataset(
    dataset_name: str,
    root: str = DATA_DIR,
    split: Optional[str] = "train",
    transform: Optional[Callable] = None,
    download: Optional[bool] = True,
    **kwargs,
):
    """Create dataset factory.

    Args:
        dataset_name (string): Name of the dataset.
            Already implemented:
                `cifar10`, `cifar100`, `stl10`, `svhn`, `mnist`, `fashion_mnist`,
                `kmnist`, `emnist`, `mnist_c`, `english_chars`, `isun`, `lsun_c`, `lsun_r`,
                `tiny_imagenet_c`, `tiny_imagenet_r`, `tiny_imagenet`, `textures`, `gaussian`,
                `uniform`, `places365`, `stanford_cars`, `imagenet`, `imagenet1k`, `ilsvrc2012`,
                `mos_inaturalist`, `mos_places365`, `mos_sun`, `cifar10_lt`, `cifar100_lt`,
                `imagenet1k_lt`, `cifar10_c`, `cifar100_c`, `imagenet_c`, `imagenet_c_npz`,
                `imagenet_a`, `imagenet_r`, `imagenet_o`, `openimage_o`, `oxford_pets`,
                `oxford_flowers`, `cub200`, `imagenet1k_c`, `blobs`, `rademacher`,
                `wilds_iwildcam`, `wilds_fmow`, `wilds_camelyon17`, `wilds_rxrx1`,
                `wilds_poverty`, `wilds_globalwheat`.
        root (string): Root directory of dataset.
        split (string, optional): Depends on the selected dataset.
        transform (callable, optional): A function/transform that  takes in an PIL image
            and returns a transformed version. E.g, `transforms.RandomCrop`
        download (bool, optional): If true, downloads the dataset from the internet and
            puts it in root directory. If dataset is already downloaded, it is not
            downloaded again.
        **kwargs: Additional arguments for dataset.

    Raises:
        ValueError: If dataset name is not specified.

    Returns:
        Dataset: Dataset object.
    """
    try:
        if dataset_name in ["imagenet", "imagenet1k", "ilsvrc2012"]:
            return datasets_registry[dataset_name](root=IMAGENET_ROOT, split=split, transform=transform, **kwargs)
        return datasets_registry[dataset_name](root=root, split=split, transform=transform, download=download, **kwargs)
    except KeyError as e:
        _logger.error(e)
        raise ValueError("Dataset name is not specified")


def get_dataset_cls(dataset_name: str) -> Type[Dataset]:
    """Return dataset class by name.

    Args:
        dataset_name (string): Name of the dataset.

    Raises:
        ValueError: If dataset name is not available in `datasets_registry`.

    Returns:
        Dataset: Dataset class.
    """
    return datasets_registry[dataset_name]


def list_datasets() -> List[str]:
    """List of available dataset names, sorted alphabetically.

    Returns:
        list: List of available dataset names.
    """
    return sorted(list(k for k in datasets_registry.keys() if datasets_registry[k] is not None))


DatasetsRegistry = Enum("DatasetsRegistry", dict(zip(list_datasets(), list_datasets())))



import logging
from functools import reduce
from typing import List, Optional

import torch
import torch.utils.data
from torch import Tensor

_logger = logging.getLogger(__name__)

HYPERPARAMETERS = dict(p=dict(low=0.5, high=0.95, step=0.05))


def get_composed_attr(model, attrs: List[str]):
    return reduce(lambda x, y: getattr(x, y), attrs, model)


class Dice:
    """
    DICE: Leveraging Sparsification for Out-of-Distribution Detection

    Args:
        model (torch.nn.Module)
        last_layer_name (Optional[str]): Name of the last layer of the model. If None, it will be inferred from the model's
            default_cfg.
        p (float): Percentage of nodes to keep in the last layer. Default: 0.7

    References:
        - Paper: https://doi.org/10.48550/arXiv.2111.09805
    """

    def __init__(self, model: torch.nn.Module, last_layer_name: Optional[str] = None, p=0.7, **kwargs) -> None:
        self.model = model
        self.p = p

        self.last_layer_name = last_layer_name
        if self.last_layer_name is None:
            if hasattr(self.model, "default_cfg"):
                self.last_layer_name = self.model.default_cfg["classifier"]
            else:
                self.last_layer_name = list(model._modules.keys())[-1]

        self._weight_backup = get_composed_attr(self.model, self.last_layer_name.split(".")).weight.clone()
        self._bias_backup = get_composed_attr(self.model, self.last_layer_name.split(".")).bias.clone()

        self.last_layer_nodes = self.last_layer_name.split(".")

        weight = get_composed_attr(self.model, self.last_layer_nodes).weight.clone()

        self.m = weight.shape[1]
        self.top_k = int(self.m * (1 - self.p))
        _logger.info("Dice top_k: %s ", self.top_k)

        self.mask = torch.ones_like(weight)
        top_k_weights = torch.topk(weight, self.top_k, dim=1).values
        self.mask[weight <= top_k_weights[:, -1].unsqueeze(1)] = 0

        get_composed_attr(self.model, self.last_layer_nodes).weight.data *= self.mask.data
        _logger.info(
            (get_composed_attr(self.model, self.last_layer_nodes).weight.data - self._weight_backup.data).sum().item()
        )
        # assert not torch.allclose(
        #     get_composed_attr(self.model, self.last_layer_nodes).weight.data, self._weight_backup.data
        # )
        # assert torch.allclose(get_composed_attr(self.model, self.last_layer_nodes).bias.data, self._bias_backup.data)

    @torch.no_grad()
    def __call__(self, x: Tensor) -> Tensor:
        logits = self.model(x)
        return torch.logsumexp(logits, dim=-1)



from functools import partial
from typing import Callable, Dict, List

import torch
import torch.fx as fx
from torchvision.models.feature_extraction import get_graph_node_names

from detectors.methods.projection import Projection
from detectors.methods.react import condition_fn, insert_fn, reactify


class ReActProjection(Projection):
    def __init__(
        self,
        model: torch.nn.Module,
        features_nodes: List[str],
        pooling_name: str = "max",
        graph_nodes_names_thr: Dict[str, float] = {"flatten": 1.0},
        insert_node_fn: Callable = insert_fn,
        aggregation_method=None,
        *args,
        **kwargs,
    ):
        self.graph_nodes_names_thr = graph_nodes_names_thr
        self.insert_node_fn = insert_node_fn
        for node_name, thr in self.graph_nodes_names_thr.items():
            model = reactify(
                model,
                condition_fn=partial(condition_fn, equals_to=node_name),
                insert_fn=partial(self.insert_node_fn, thr=thr),
            )

        super().__init__(model, features_nodes, pooling_name, aggregation_method, *args, **kwargs)


def test():
    import torchvision.models as models

    model = models.densenet121()
    print(get_graph_node_names(model)[0])
    graph: fx.Graph = fx.Tracer().trace(model)
    model = fx.GraphModule(model, graph)
    print(model.code)
    graph_nodes_names_thr = {
        "features_transition1_pool": 1.0,
        "features_transition2_pool": 1.0,
        "features_transition3_pool": 1,
        "features_norm5": 1,
        "flatten": 1,
    }
    for node_name, thr in graph_nodes_names_thr.items():
        model = reactify(model, partial(condition_fn, equals_to=node_name), partial(insert_fn, thr=thr))
    print(model.graph)
    print(model.code)

    model.classifier = torch.nn.Linear(1024, 3)
    model.eval()
    x = torch.rand(32, 3, 224, 224)
    y = torch.randint(0, 3, (32,))
    projection = ReActProjection(
        model, ["clip", "clip_1", "clip_2", "clip_3", "clip_4", "classifier"], "max", graph_nodes_names_thr
    )
    projection.fit(x, y)
    projection.end()
    print(projection(x))
    assert projection(x).shape == (32,)


if __name__ == "__main__":
    test()



"""ResNet models for CIFAR10, CIFAR100, and SVHN datasets."""
import logging

import timm
import timm.models
import torch
import torch.nn as nn
from timm.models import register_model as timm_register_model

from detectors.data import CIFAR10_DEFAULT_MEAN, CIFAR10_DEFAULT_STD
from detectors.data.constants import CIFAR100_DEFAULT_MEAN, CIFAR100_DEFAULT_STD, SVHN_DEFAULT_MEAN, SVHN_DEFAULT_STD
from detectors.models.utils import ModelDefaultConfig, hf_hub_url_template

_logger = logging.getLogger(__name__)


def _cfg(url="", **kwargs):
    num_classes = kwargs.pop("num_classes", 10)
    mean = kwargs.pop("mean", CIFAR10_DEFAULT_MEAN)
    std = kwargs.pop("std", CIFAR10_DEFAULT_STD)
    return ModelDefaultConfig(
        url=url,
        num_classes=num_classes,
        input_size=(3, 32, 32),
        pool_size=(4, 4),
        crop_pct=1,
        interpolation="bilinear",
        mean=mean,
        std=std,
        first_conv="conv1",
        classifier="fc",
        **kwargs,
    )


default_cfgs = {
    # CIFAR-10
    "resnet18_cifar10": _cfg(url=hf_hub_url_template("resnet18_cifar10"), architecture="resnet18"),
    "resnet34_cifar10": _cfg(url=hf_hub_url_template("resnet34_cifar10"), architecture="resnet34"),
    "resnet50_cifar10": _cfg(url=hf_hub_url_template("resnet50_cifar10"), architecture="resnet50"),
    "resnet34_simclr_cifar10": _cfg(
        url=hf_hub_url_template("resnet34_simclr_cifar10"), architecture="resnet34", num_classes=512
    ),
    "resnet50_simclr_cifar10": _cfg(
        url=hf_hub_url_template("resnet50_simclr_cifar10"), architecture="resnet50", num_classes=2048
    ),
    "resnet34_supcon_cifar10": _cfg(
        url=hf_hub_url_template("resnet34_supcon_cifar10"), architecture="resnet34", num_classes=512
    ),
    "resnet50_supcon_cifar10": _cfg(
        url=hf_hub_url_template("resnet50_supcon_cifar10"), architecture="resnet50", num_classes=2048
    ),
    # CIFAR-100
    "resnet18_cifar100": _cfg(
        url=hf_hub_url_template("resnet18_cifar100"),
        num_classes=100,
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        architecture="resnet18",
    ),
    "resnet34_cifar100": _cfg(
        url=hf_hub_url_template("resnet34_cifar100"),
        num_classes=100,
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        architecture="resnet34",
    ),
    "resnet50_cifar100": _cfg(
        url=hf_hub_url_template("resnet50_cifar100"),
        num_classes=100,
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        architecture="resnet50",
    ),
    "resnet34_simclr_cifar100": _cfg(
        url=hf_hub_url_template("resnet34_simclr_cifar100"),
        architecture="resnet34",
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        num_classes=512,
    ),
    "resnet50_simclr_cifar100": _cfg(
        url=hf_hub_url_template("resnet50_simclr_cifar100"),
        architecture="resnet50",
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        num_classes=2048,
    ),
    "resnet34_supcon_cifar100": _cfg(
        url=hf_hub_url_template("resnet34_supcon_cifar100"),
        architecture="resnet34",
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        num_classes=512,
    ),
    "resnet50_supcon_cifar100": _cfg(
        url=hf_hub_url_template("resnet50_supcon_cifar100"),
        architecture="resnet50",
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        num_classes=2048,
    ),
    # SVHN
    "resnet18_svhn": _cfg(
        url=hf_hub_url_template("resnet18_svhn"), mean=SVHN_DEFAULT_MEAN, std=SVHN_DEFAULT_STD, architecture="resnet18"
    ),
    "resnet34_svhn": _cfg(
        url=hf_hub_url_template("resnet34_svhn"), mean=SVHN_DEFAULT_MEAN, std=SVHN_DEFAULT_STD, architecture="resnet34"
    ),
    "resnet50_svhn": _cfg(
        url=hf_hub_url_template("resnet50_svhn"), mean=SVHN_DEFAULT_MEAN, std=SVHN_DEFAULT_STD, architecture="resnet50"
    ),
}


def _create_resnet_small(variant, features_dim=512, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    architecture = default_cfg.architecture or variant.split("_")[0]
    model = timm.create_model(architecture, pretrained=False)

    # override timm config
    model.default_cfg = default_cfg
    model.pretrained_cfg = default_cfg

    # override model
    model.conv1 = nn.Conv2d(3, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)
    model.maxpool = nn.Identity()  # type: ignore
    model.fc = nn.Linear(features_dim, model.default_cfg.num_classes)

    # load weights
    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(model.default_cfg.url, map_location="cpu", file_name=f"{variant}.pth")
        )

    return model


def _create_resnet_small_ssl(variant, features_dim=512, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    architecture = default_cfg.architecture or variant.split("_")[0]
    model = timm.create_model(architecture, pretrained=False, num_classes=0)

    # override timm config
    model.default_cfg = default_cfg
    model.pretrained_cfg = default_cfg

    # override model
    model.conv1 = nn.Conv2d(3, 64, kernel_size=(3, 3), stride=(1, 1), padding=(1, 1), bias=False)
    model.maxpool = nn.Identity()  # type: ignore

    if pretrained:
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(model.default_cfg.url, map_location="cpu", file_name=f"{variant}.pth")
        )
    return model


@timm_register_model
def resnet18_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small("resnet18_cifar10", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small("resnet34_cifar10", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small("resnet50_cifar10", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_simclr_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet34_simclr_cifar10", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_simclr_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet50_simclr_cifar10", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_supcon_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet34_supcon_cifar10", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_supcon_cifar10(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet50_supcon_cifar10", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet18_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small("resnet18_cifar100", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small("resnet34_cifar100", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small("resnet50_cifar100", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_simclr_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet34_simclr_cifar100", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_simclr_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet50_simclr_cifar100", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_supcon_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet34_supcon_cifar100", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_supcon_cifar100(pretrained=False, **kwargs):
    return _create_resnet_small_ssl("resnet50_supcon_cifar100", features_dim=2048, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet18_svhn(pretrained=False, **kwargs):
    return _create_resnet_small("resnet18_svhn", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet34_svhn(pretrained=False, **kwargs):
    return _create_resnet_small("resnet34_svhn", features_dim=512, pretrained=pretrained, **kwargs)


@timm_register_model
def resnet50_svhn(pretrained=False, **kwargs):
    return _create_resnet_small("resnet50_svhn", features_dim=2048, pretrained=pretrained, **kwargs)



import logging
import os
from typing import Callable, Optional

import numpy as np
from PIL import Image
from torch.utils.data.dataset import Dataset
from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive, verify_str_arg
from tqdm import tqdm

_logger = logging.getLogger(__name__)


class ImageNetA(ImageFolder):
    """ImageNetA dataset.

    - Paper: [https://arxiv.org/abs/1907.07174](https://arxiv.org/abs/1907.07174).
    """

    base_folder = "imagenet-a"
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-a.tar"
    filename = "imagenet-a.tar"
    tgz_md5 = "c3e55429088dc681f30d81f4726b6595"

    def __init__(self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs):

        self.root = root

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        super().__init__(root=os.path.join(root, self.base_folder), transform=transform, **kwargs)

    def _check_exists(self) -> bool:
        return os.path.exists(os.path.join(self.root, self.base_folder))

    def _check_integrity(self) -> bool:
        return check_integrity(os.path.join(self.root, self.filename), self.tgz_md5)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            _logger.debug("Files already downloaded and verified")
            return
        download_and_extract_archive(self.url, self.root, filename=self.filename, md5=self.tgz_md5)


class ImageNetO(ImageNetA):
    """ImageNetO datasets.

    Contains unknown classes to ImageNet-1k.


    - Paper: [https://arxiv.org/abs/1907.07174](https://arxiv.org/abs/1907.07174)
    """

    base_folder = "imagenet-o"
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-o.tar"
    filename = "imagenet-o.tar"
    tgz_md5 = "86bd7a50c1c4074fb18fc5f219d6d50b"


class ImageNetR(ImageNetA):
    """ImageNet-R(endition) dataset.

    Contains art, cartoons, deviantart, graffiti, embroidery, graphics, origami, paintings,
    patterns, plastic objects,plush objects, sculptures, sketches, tattoos, toys,
    and video game renditions of ImageNet-1k classes.

    - Paper: [https://arxiv.org/abs/2006.16241](https://arxiv.org/abs/2006.16241)
    """

    base_folder = "imagenet-r"
    url = "https://people.eecs.berkeley.edu/~hendrycks/imagenet-r.tar"
    filename = "imagenet-r.tar"
    tgz_md5 = "a61312130a589d0ca1a8fca1f2bd3337"


CORRUPTIONS = [
    "brightness",
    "contrast",
    "defocus_blur",
    "elastic_transform",
    "fog",
    "frost",
    "gaussian_blur",
    "gaussian_noise",
    "glass_blur",
    "impulse_noise",
    "jpeg_compression",
    "motion_blur",
    "pixelate",
    "saturate",
    "shot_noise",
    "snow",
    "spatter",
    "speckle_noise",
    "zoom_blur",
]


class ImageNetC(ImageNetA):
    """Corrupted version of the ImageNet-1k dataset.

    It contains the following subsets:

    - `noise` (21GB): gaussian_noise, shot_noise, and impulse_noise.
    - `blur` (7GB): defocus_blur, glass_blur, motion_blur, and zoom_blur.
    - `weather` (12GB):  frost, snow, fog, and brightness.
    - `digital` (7GB): contrast, elastic_transform, pixelate, and jpeg_compression.
    - `extra` (15GB): speckle_noise, spatter, gaussian_blur, and saturate.

    - Paper: [https://arxiv.org/abs/1903.12261v1](https://arxiv.org/abs/1903.12261v1)
    """

    split_list = ["blur", "digital", "extra", "noise", "weather"]
    base_folder_name = "ImageNetC"
    url_base = "https://zenodo.org/record/2235448/files/"
    tgz_md5_list = [
        "2d8e81fdd8e07fef67b9334fa635e45c",
        "89157860d7b10d5797849337ca2e5c03",
        "d492dfba5fc162d8ec2c3cd8ee672984",
        "e80562d7f6c3f8834afb1ecf27252745",
        "33ffea4db4d93fe4a428c40a6ce0c25d",
    ]
    corruptions = CORRUPTIONS

    def __init__(
        self,
        root: str,
        split: str,
        intensity: int,
        transform: Optional[Callable] = None,
        download: bool = False,
        **kwargs,
    ) -> None:
        self.root = os.path.expanduser(root)
        self.corruption = verify_str_arg(split, "split", self.corruptions)
        split_group = self._get_corruption_group(self.corruption)

        self._base_folder = os.path.join(root, self.base_folder_name, split_group)
        self.filename = split_group + ".tar"
        self.url = self.url_base + self.filename
        self.tgz_md5 = self.tgz_md5_list[self.split_list.index(split_group)]

        self.base_folder = os.path.join(self._base_folder, split, str(intensity))

        super().__init__(root, transform=transform, download=download, **kwargs)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            _logger.debug("Files already downloaded and verified")
            return
        download_and_extract_archive(
            self.url, self.root, extract_root=self._base_folder, filename=self.filename, md5=self.tgz_md5
        )

    @staticmethod
    def _get_corruption_group(corruption: str):
        split_group = ""
        if corruption in ["defocus_blur", "glass_blur", "motion_blur", "zoom_blur"]:
            split_group = "blur"
        elif corruption in ["contrast", "elastic_transform", "pixelate", "jpeg_compression"]:
            split_group = "digital"
        elif corruption in ["speckle_noise", "spatter", "gaussian_blur", "saturate"]:
            split_group = "extra"
        elif corruption in ["gaussian_noise", "shot_noise", "impulse_noise"]:
            split_group = "noise"
        elif corruption in ["frost", "snow", "fog", "brightness"]:
            split_group = "weather"
        return split_group


def _imagenet_c_to_npz(root: str, split: str, intensity: int, dest_folder: str = "ImageNetCnpz") -> None:

    dataset = ImageNetC(root, split, intensity, download=True)
    assert len(dataset) == 50_000, "ImageNetC should have 50,000 images. Please check the dataset."
    image_example = dataset[0][0]
    width, height = image_example.size
    _logger.info("Image size: %d x %d", width, height)
    x = np.ndarray(shape=(len(dataset), height, width, 3), dtype=np.uint8)
    y = np.ndarray(shape=(len(dataset)), dtype=np.int32)
    for i in tqdm(range(len(dataset))):
        image, label = dataset[i]
        x[i] = image
        y[i] = label

    os.makedirs(os.path.join(root, dest_folder), exist_ok=True)
    np.savez(os.path.join(root, dest_folder, f"{split}-{intensity}.npz"), x=x, y=y)


class ImageNetCnpz(Dataset):
    """Corrupted version of the ImageNet-1k dataset saved in npz format."""

    corruptions = CORRUPTIONS
    base_folder_name = "ImageNetCnpz"

    def __init__(
        self,
        root: str,
        split: str,
        intensity: int,
        transform: Optional[Callable] = None,
        download: bool = False,
        **kwargs,
    ) -> None:
        super().__init__()
        self.root = os.path.expanduser(root)
        self.corruption = verify_str_arg(split, "split", self.corruptions)
        self.intensity = int(intensity)
        self.path = os.path.join(self.root, self.base_folder_name, f"{split}-{intensity}.npz")
        self.transform = transform
        if download:
            self.download()

        data = np.load(self.path, mmap_mode="r")
        self.images = data["x"]
        self.labels = data["y"]

    def __getitem__(self, index):
        x = self.images[index]
        x = Image.fromarray(x)

        if self.transform:
            x = self.transform(x)

        y = self.labels[index]
        return x, y

    def __len__(self):
        return len(self.images)

    def _check_exists(self) -> bool:
        return os.path.exists(self.path)

    def download(self) -> None:
        if self._check_exists():
            return
        _imagenet_c_to_npz(self.root, self.corruption, self.intensity, self.base_folder_name)



import logging
from typing import List, Literal, Optional

import torch
from torch import Tensor, nn

from detectors.methods.gmm_torch import GaussianMixture
from detectors.methods.templates import DetectorWithFeatureExtraction

_logger = logging.getLogger(__name__)

HYPERPARAMETERS = dict(covariance_type=["full", "tied", "diag"], n_components=[None, 16, 128, 256])


class GMM(DetectorWithFeatureExtraction):
    def __init__(
        self,
        model: nn.Module,
        features_nodes: Optional[List[str]] = None,
        all_blocks: bool = False,
        last_layer: bool = False,
        pooling_op_name: str = "avg_or_getitem",
        aggregation_method_name: str = "mean",
        n_components: Optional[int] = None,
        covariance_type: Literal["full", "tied", "diag"] = "full",
        **kwargs_gmm
    ):
        super().__init__(
            model,
            features_nodes=features_nodes,
            all_blocks=all_blocks,
            last_layer=last_layer,
            pooling_op_name=pooling_op_name,
            aggregation_method_name=aggregation_method_name,
        )
        self.n_components = n_components
        self.covariance_type = covariance_type
        self.kwargs_gmm = kwargs_gmm

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        return self.gms[layer_name].score_samples(x).view(-1)

    def _fit_params(self) -> None:
        _logger.info("Estimating GMM parameters...")

        # estimate GMM parameters
        if self.n_components is None:
            self.n_components = torch.unique(self.train_targets).shape[0]
        _logger.info("Number of components set to %i.", self.n_components)

        self.gms = {}
        device = next(self.model.parameters()).device
        for layer_name, layer_features in self.train_features.items():
            self.gms[layer_name] = GaussianMixture(
                n_components=self.n_components,
                covariance_type=self.covariance_type,
                init_params="random_from_data",
                **self.kwargs_gmm
            )
            self.gms[layer_name].fit(layer_features.to(device))



import logging
from typing import List, Optional, Union

import torch
from torch import Tensor
from tqdm import tqdm

from detectors.methods.templates import DetectorWithFeatureExtraction

_logger = logging.getLogger(__name__)


def projection_layer_score(x: Tensor, mus: Union[Tensor, List[Tensor]], eps=1e-7):
    if isinstance(mus, list):
        mus = torch.cat(mus, dim=0)
    den = torch.norm(x, dim=-1, keepdim=True) @ torch.norm(mus, dim=-1, keepdim=True).T
    stack = x @ mus.T
    stack = stack / (den + eps)
    return torch.norm(x, p=2, dim=-1, keepdim=True) * stack  # type: ignore


class Projection(DetectorWithFeatureExtraction):
    def __init__(
        self,
        model: torch.nn.Module,
        features_nodes: Optional[List[str]] = None,
        pooling_op_name: str = "max_or_getitem",
        **kwargs
    ):
        super().__init__(
            model=model,
            features_nodes=features_nodes,
            pooling_op_name=pooling_op_name,
            all_blocks=True,
            last_layer=True,
            aggregation_method_name="innerprod",
            **kwargs
        )

        self.mus = None

    def end(self, *args, **kwargs):
        for node_name, v in self.train_features.items():
            if isinstance(v, list):
                self.train_features[node_name] = torch.cat(v, dim=0)
        if isinstance(self.train_targets, list):
            self.train_targets = torch.cat(self.train_targets, dim=0)

        self._fit_params()

        if self.aggregation_method is not None and hasattr(self.aggregation_method, "fit"):
            _logger.info("Fitting aggregator %s...", self.aggregation_method_name)
            self.batch_size = self.train_targets.shape[0]  # type: ignore
            all_scores = torch.zeros(self.train_targets.shape[0], len(self.train_features))
            train_probs = torch.softmax(self.train_features[list(self.train_features.keys())[-1]], 1)
            for i, (k, v) in tqdm(enumerate(self.train_features.items())):
                idx = 0
                for idx in range(0, v.shape[0], self.batch_size):
                    all_scores[:, i] = self._layer_score(
                        v[idx : idx + self.batch_size], train_probs[idx : idx + self.batch_size], k, i
                    ).view(-1)
            self.aggregation_method.fit(all_scores, self.train_targets)

        del self.train_features
        del self.train_targets

    @torch.no_grad()
    def __call__(self, x: Tensor) -> Tensor:
        features = self.feature_extractor(x)
        logits = features[list(features.keys())[-1]]
        probs = torch.softmax(logits, dim=1)
        all_scores = torch.zeros(x.shape[0], len(features), device=x.device)
        for i, (k, v) in enumerate(features.items()):
            all_scores[:, i] = self._layer_score(v, probs, k, i).view(-1)

        all_scores = self.aggregation_method(all_scores)
        return all_scores.view(-1)

    def _layer_score(self, x: Tensor, probs: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        scores = projection_layer_score(x, self.mus[layer_name].to(x.device))  # type: ignore
        scores = torch.sum(scores * probs, 1, keepdim=True)
        return scores

    def _fit_params(self) -> None:
        self.mus = {}
        unique_classes = torch.unique(self.train_targets).detach().cpu().numpy().tolist()
        for layer_name, layer_features in self.train_features.items():
            self.mus[layer_name] = []
            for c in unique_classes:
                filt = self.train_targets == c
                if filt.sum() == 0:
                    continue
                self.mus[layer_name].append(layer_features[filt].mean(0, keepdim=True))
            self.mus[layer_name] = torch.cat(self.mus[layer_name], dim=0)



import timm
import torch
from timm.models import register_model as timm_register_model

from detectors.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from detectors.models.utils import ModelDefaultConfig


# def _cfg(url="", **kwargs):
#     model = timm.create_model(architecture, pretrained=False, num_classes=0)
#     base_cfg = model.default_cfg
#     num_classes = kwargs.pop("num_classes", 10)
#     # replace base cfg with new arguments
#     base_cfg.update(kwargs)
#     base_cfg["architecture"] = architecture
#     base_cfg["url"] = url
#     base_cfg["num_classes"] = num_classes
#     return ModelDefaultConfig(**base_cfg)


def _create_dinov2(variant: str):
    model = torch.hub.load("facebookresearch/dinov2", variant)
    return model


@timm_register_model
def dinov2_vitb14(pretrained=True, **kwargs):
    return _create_dinov2("dinov2_vitb14")



# TODO
"""Typical Feature Estimated Method (TFEM)"""
# import logging
# from functools import partial
# from typing import List, Literal, Optional

# import numpy as np
# import torch
# from torch import Tensor, nn
# from torchvision.models.feature_extraction import create_feature_extractor
# from tqdm import tqdm

# from detectors.aggregations import create_aggregation

# from .utils import create_reduction

# _logger = logging.getLogger(__name__)


# import torch


# class BATS:
#     def __init__(self, lbd=1) -> None:
#         self.lbd = lbd

#     def start(self):
#         return

#     def update(self, x: Tensor, y: Tensor):
#         return

#     def end(self):
#         return

#     def __call__(self, x: Tensor) -> Tensor:
#         return torch.ones(x.shape[0])



import timm
import timm.data
import torch
from timm import create_model, list_models
from timm.models import register_model as timm_register_model

from .densenet import *
from .resnet import *
from .utils import *
from .vgg import *
from .vit import *


def create_transform(model: torch.nn.Module, is_training: bool = False):
    """Create a input transformation for a given model.

    Based on the default configuration of the model following timm's library.

    Args:
        model (torch.nn.Module): Model to create the transformation for.
        is_training (bool, optional): Whether the transformation is for training or not. Defaults to False.

    Returns:
        Callable: The transformation.
    """
    data_config = timm.data.resolve_data_config(model.default_cfg)
    data_config["is_training"] = is_training
    transform = timm.data.create_transform(**data_config)
    return transform



"""VGG models for CIFAR10, CIFAR100 and SVHN datasets."""
import timm
import timm.models
import torch
import torch.nn as nn
from timm.models import register_model as timm_register_model

from detectors.data import CIFAR10_DEFAULT_MEAN, CIFAR10_DEFAULT_STD
from detectors.data.constants import CIFAR100_DEFAULT_MEAN, CIFAR100_DEFAULT_STD, SVHN_DEFAULT_MEAN, SVHN_DEFAULT_STD
from detectors.models.utils import ModelDefaultConfig, hf_hub_url_template


def _cfg(url="", **kwargs):
    num_classes = kwargs.pop("num_classes", 10)
    mean = kwargs.pop("mean", CIFAR10_DEFAULT_MEAN)
    std = kwargs.pop("std", CIFAR10_DEFAULT_STD)
    return ModelDefaultConfig(
        url=url,
        num_classes=num_classes,
        input_size=(3, 32, 32),
        pool_size=(4, 4),
        crop_pct=1,
        interpolation="bilinear",
        mean=mean,
        std=std,
        first_conv="features.0",
        classifier="head.fc",
        **kwargs,
    )


default_cfgs = {
    "vgg16_bn_cifar10": _cfg(
        url=hf_hub_url_template("vgg16_bn_cifar10"),
        architecture="vgg16_bn",
    ),
    "vgg16_bn_cifar100": _cfg(
        url=hf_hub_url_template("vgg16_bn_cifar100"),
        num_classes=100,
        mean=CIFAR100_DEFAULT_MEAN,
        std=CIFAR100_DEFAULT_STD,
        architecture="vgg16_bn",
    ),
    "vgg16_bn_svhn": _cfg(
        url=hf_hub_url_template("vgg16_bn_svhn"),
        mean=SVHN_DEFAULT_MEAN,
        std=SVHN_DEFAULT_STD,
        architecture="vgg16_bn",
    ),
}


def _create_vgg_small(variant, features_dim=512, pretrained=False, **kwargs):
    default_cfg = default_cfgs[variant]

    # load timm model
    architecture = default_cfg.architecture
    model = timm.create_model(architecture, pretrained=False)

    # override timm config
    model.default_cfg = default_cfg
    model.pretrained_cfg = default_cfg

    # override model
    model.pre_logits = nn.Identity()  # type: ignore
    model.head.fc = nn.Linear(features_dim, model.default_cfg.num_classes)

    # load weights
    if pretrained:
        checkpoint = model.default_cfg.url
        model.load_state_dict(
            torch.hub.load_state_dict_from_url(
                checkpoint, progress=True, map_location="cpu", file_name=f"{variant}.pth"
            )
        )

    return model


@timm_register_model
def vgg16_bn_cifar10(pretrained=False, **kwargs):
    return _create_vgg_small("vgg16_bn_cifar10", pretrained=pretrained, **kwargs)


@timm_register_model
def vgg16_bn_cifar100(pretrained=False, **kwargs):
    return _create_vgg_small("vgg16_bn_cifar100", pretrained=pretrained, **kwargs)


@timm_register_model
def vgg16_bn_svhn(pretrained=False, **kwargs):
    return _create_vgg_small("vgg16_bn_svhn", pretrained=pretrained, **kwargs)



from typing import Optional

import torch
from torch import Tensor, nn

from detectors.methods.templates import DetectorWithFeatureExtraction


class CSI(DetectorWithFeatureExtraction):
    """CSI: Novelty Detection via Contrastive Learning on Distributionally Shifted Instances

    Extract features from the last layer of a self supervised model.

    References:
        [1] https://arxiv.org/abs/2007.08176
        [2] https://github.com/alinlab/CSI
    """

    def __init__(self, model: nn.Module, **kwargs) -> None:
        super().__init__(
            model,
            features_nodes=[list(model._modules.keys())[-1]],
            all_blocks=False,
            last_layer=False,
            pooling_op_name="none",
            aggregation_method_name="none",
        )

    def _fit_params(self) -> None:
        self.nearest_neighbors = self.train_features

    def _layer_score(self, features: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None, **kwargs):
        den = torch.norm(features, dim=-1, keepdim=True) * torch.norm(
            self.nearest_neighbors[layer_name], dim=-1, keepdim=True
        )
        num = features @ self.nearest_neighbors[layer_name].T
        stack = num / (den + 1e-7)
        return torch.norm(features, p=2, dim=-1) * torch.max(stack, dim=-1)[0]  # type: ignore



import os
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class EnglishChars(ImageFolder):
    """In the English language, Latin script (excluding accents) and Hindu-Arabic numerals are used.
    For simplicity we call this the "English" characters set. The dataset consists of:

    * 64 classes (0-9, A-Z, a-z)
    * 7705 characters obtained from natural images
    * 3410 hand drawn characters using a tablet PC
    * 62992 synthesised characters from computer fonts
    * This gives a total of over 74K images (which explains the name of the dataset).

    """

    base_folder = "chars74k"
    images_folder = "English/Img/GoodImg/Bmp/"
    filename = "EnglishImg.tgz"
    file_md5 = "85d157e0c58f998e1cda8def62bcda0d"
    url = "http://www.ee.surrey.ac.uk/CVSSP/demos/chars74k/EnglishImg.tgz"

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")
        super().__init__(self._split_folder, transform=transform, **kwargs)

    @property
    def _dataset_folder(self):
        return os.path.join(self.root, self.base_folder)

    @property
    def _split_folder(self):
        return os.path.join(self._dataset_folder, self.images_folder)

    def _check_integrity(self) -> bool:
        root = self.root
        md5 = self.file_md5
        fpath = os.path.join(root, self.filename)
        return check_integrity(fpath, md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self._split_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        download_and_extract_archive(
            self.url, download_root=self.root, extract_root=self._dataset_folder, md5=self.file_md5
        )



import os
from typing import Callable, Optional

from torchvision.datasets import ImageFolder
from torchvision.datasets.utils import check_integrity, download_and_extract_archive


class Textures(ImageFolder):
    """DTD is a texture database, consisting of 5640 images, organized according
    to a list of 47 terms (categories) inspired from human perception. There are 120 images
    for each category. Image sizes range between 300x300 and 640x640, and the images
    contain at least 90% of the surface representing the category attribute. The images
    were collected from Google and Flickr by entering our proposed attributes and related
    terms as search queries."""

    base_folder = "textures"
    images_folder = "dtd/images"
    filename = "dtd-r1.0.1.tar.gz"
    file_md5 = "fff73e5086ae6bdbea199a49dfb8a4c1"
    url = "https://www.robots.ox.ac.uk/~vgg/data/dtd/download/dtd-r1.0.1.tar.gz"

    def __init__(
        self, root: str, split=None, transform: Optional[Callable] = None, download: bool = False, **kwargs
    ) -> None:
        self.root = os.path.expanduser(root)

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError("Dataset not found or corrupted." + " You can use download=True to download it")

        super().__init__(self.split_folder, transform=transform, **kwargs)

    @property
    def dataset_folder(self):
        return os.path.join(self.root, self.base_folder)

    @property
    def split_folder(self):
        return os.path.join(self.dataset_folder, self.images_folder)

    def _check_integrity(self) -> bool:
        root = self.root
        md5 = self.file_md5
        fpath = os.path.join(root, self.filename)
        return check_integrity(fpath, md5)

    def _check_exists(self) -> bool:
        return os.path.exists(self.split_folder)

    def download(self) -> None:
        if self._check_integrity() and self._check_exists():
            return
        download_and_extract_archive(
            self.url, download_root=self.root, extract_root=self.dataset_folder, md5=self.file_md5
        )



"""
Adapted from https://github.com/Megvii-Nanjing/BBN
"""

from typing import Callable, Literal, Optional

import numpy as np
import torchvision
from PIL import Image

# CIFAR10:
# many: 0,1,2
# median: 3,4,5,6
# few: 7,8,9


class CIFAR10LT(torchvision.datasets.CIFAR10):
    cls_num = 10

    def __init__(
        self,
        root: str,
        train: bool,
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        download=True,
        imbalance_ratio=0.01,
        imb_type: Literal["exp", "step"] = "exp",
    ):
        super().__init__(root, train, transform=transform, target_transform=target_transform, download=download)
        self.train = train
        if self.train:
            img_num_list = self._get_img_num_per_cls(self.cls_num, imb_type, imbalance_ratio)
            self._gen_imbalanced_data(img_num_list)

        self.labels = self.targets

    def _get_img_num_per_cls(self, cls_num, imb_type, imb_factor):
        img_max = len(self.data) / cls_num
        img_num_per_cls = []
        if imb_type == "exp":
            for cls_idx in range(cls_num):
                num = img_max * (imb_factor ** (cls_idx / (cls_num - 1.0)))
                img_num_per_cls.append(int(num))
        elif imb_type == "step":
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max))
            for cls_idx in range(cls_num // 2):
                img_num_per_cls.append(int(img_max * imb_factor))
        else:
            img_num_per_cls.extend([int(img_max)] * cls_num)
        self.img_num_per_cls = img_num_per_cls
        return img_num_per_cls

    def _gen_imbalanced_data(self, img_num_per_cls):
        new_data = []
        new_targets = []
        targets_np = np.array(self.targets, dtype=np.int64)
        classes = np.unique(targets_np)

        self.num_per_cls_dict = dict()
        for the_class, the_img_num in zip(classes, img_num_per_cls):
            self.num_per_cls_dict[the_class] = the_img_num
            idx = np.where(targets_np == the_class)[0]
            # np.random.shuffle(idx)
            selec_idx = idx[:the_img_num]
            new_data.append(self.data[selec_idx, ...])
            new_targets.extend([the_class] * the_img_num)
        new_data = np.vstack(new_data)
        self.data = new_data
        self.targets = new_targets

    def __getitem__(self, index):
        img, label = self.data[index], self.labels[index]
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            label = self.target_transform(label)

        return img, label

    def __len__(self):
        return len(self.labels)


class CIFAR100LT(CIFAR10LT):
    cls_num = 100
    base_folder = "cifar-100-python"
    url = "https://www.cs.toronto.edu/~kriz/cifar-100-python.tar.gz"
    filename = "cifar-100-python.tar.gz"
    tgz_md5 = "eb9058c3a382ffc7106e4002c42a8d85"
    train_list = [
        ["train", "16019d7e3df5f24257cddd939b257f8d"],
    ]
    test_list = [
        ["test", "f0ef6b0ae62326f3e7ffdfab6717acfc"],
    ]
    meta = {
        "filename": "meta",
        "key": "fine_label_names",
        "md5": "7973b15100ade9c7d40fb424638fde48",
    }


def test():
    cifar10 = torchvision.datasets.CIFAR10(root="data", train=True, download=True)
    cifar10lt = CIFAR10LT(root="data", train=True, download=True, imb_type="exp", imbalance_ratio=0.01)
    cifar100 = torchvision.datasets.CIFAR100(root="data", train=True, download=True)
    cifar100lt = CIFAR100LT(root="data", train=True, download=True, imb_type="exp", imbalance_ratio=0.01)

    print(len(cifar10))

    print(len(cifar10lt))
    print(cifar10lt.img_num_per_cls)

    print(len(cifar100))

    print(len(cifar100lt))
    print(cifar100lt.img_num_per_cls)


if __name__ == "__main__":
    test()



import logging
import types
from functools import partial

from torch import Tensor

from .anomaly import IFAggregation, LOFAggregation
from .basics import (
    avg_topk_aggregation,
    depth_weighted_sum,
    layer_idx,
    max_aggregation,
    mean_aggregation,
    median_aggregation,
    min_aggregation,
    none_aggregation,
    topk_aggregation,
)
from .cosine import CosineAggregation
from .innerprod import InnerProductAggregation
from .mahalanobis import MahalanobisAggregation
from .quantile import QuantileAggregation

_logger = logging.getLogger(__name__)

aggregations_registry = {
    "none": none_aggregation,
    "mean": mean_aggregation,
    "max": max_aggregation,
    "min": min_aggregation,
    "median": median_aggregation,
    "dws": depth_weighted_sum,
    "avg_topk": avg_topk_aggregation,
    "topk": topk_aggregation,
    "lof": LOFAggregation,
    "if": IFAggregation,
    "layer_idx": layer_idx,
    "mahalanobis": MahalanobisAggregation,
    "innerprod": InnerProductAggregation,
    "cosine": CosineAggregation,
    "quantile": QuantileAggregation,
}


class Aggregation:
    """Aggregation wrapper class."""

    def __init__(self, aggregation_method, *args, **kwargs) -> None:
        self.aggregation_method = aggregation_method

    def fit(self, stack: Tensor, *args, **kwargs):
        if not hasattr(self.aggregation_method, "fit"):
            _logger.debug("Aggregation method does not have a `fit` method.")
            return
        self.aggregation_method.fit(stack)

    def __call__(self, stack: Tensor, *args, **kwargs):
        return self.aggregation_method(stack, *args, **kwargs)


def register_aggregation(name: str):
    """Decorator to register a new aggregation method."""

    def decorator(f):
        aggregations_registry[name] = f
        return f

    return decorator


def create_aggregation(aggregation_name: str, **kwargs) -> Aggregation:
    if aggregation_name not in aggregations_registry:
        raise ValueError(f"Unknown aggregation method: {aggregation_name}")
    if not isinstance(aggregations_registry[aggregation_name], types.FunctionType):
        return Aggregation(aggregations_registry[aggregation_name](**kwargs))
    return Aggregation(partial(aggregations_registry[aggregation_name], **kwargs))


def list_aggregations() -> list:
    return list(aggregations_registry.keys())



import logging
from typing import List, Optional, Union

import torch
from torch import Tensor

from detectors.methods.templates import DetectorWithFeatureExtraction

_logger = logging.getLogger(__name__)


def max_cosine_sim_layer_score(x: Tensor, mus: Union[Tensor, List[Tensor]], eps=1e-7):
    if isinstance(mus, list):
        mus = torch.cat(mus, dim=0)
    den = torch.norm(x, dim=-1, keepdim=True) @ torch.norm(mus, dim=-1, keepdim=True).T
    stack = x @ mus.T
    stack = stack / (den + eps)
    return stack.max(dim=-1)[0]


class MaxCosineSimilarity(DetectorWithFeatureExtraction):
    """MaxCosineSimilarity detector.


    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of:
                `max`, `avg`, `none`, `flatten`, `getitem`, `avg_or_getitem`, `max_or_getitem`.
            Defaults to "avg".
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.

    References:
        [1] https://openaccess.thecvf.com/content/ACCV2020/html/Techapanurak_Hyperparameter-Free_Out-of-Distribution_Detection_Using_Cosine_Similarity_ACCV_2020_paper.html
    """

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        return max_cosine_sim_layer_score(x, self.mus[layer_name].to(x.device))

    def _fit_params(self) -> None:
        self.mus = {}
        unique_classes = torch.unique(self.train_targets).detach().cpu().numpy().tolist()
        for layer_name, layer_features in self.train_features.items():
            self.mus[layer_name] = []
            for c in unique_classes:
                filt = self.train_targets == c
                if filt.sum() == 0:
                    continue
                self.mus[layer_name].append(layer_features[filt].mean(0, keepdim=True))
            self.mus[layer_name] = torch.cat(self.mus[layer_name], dim=0)



import logging
from typing import Optional

import numpy as np
import torch
import torch.distributed as dist
from sklearn.covariance import EmpiricalCovariance
from torch import Tensor, nn
from torchvision.models.feature_extraction import create_feature_extractor

from detectors.methods.utils import get_composed_attr
from detectors.utils import sync_tensor_across_gpus

_logger = logging.getLogger(__name__)


class ViM:
    """Virtual Logit Matching (ViM) detector.

    Args:
        model (torch.nn.Module): Model to be used to extract features
        last_layer_name (Optional[str]): Name of the last layer. Defaults to None.
        penultimate_layer_name (Optional[str]): Name of the penultimate layer. Defaults to None.

    References:
        [1] https://arxiv.org/abs/2203.10807
    """

    def __init__(
        self,
        model: nn.Module,
        last_layer_name: Optional[str] = None,
        penultimate_layer_name: Optional[str] = None,
        **kwargs
    ):
        self.model = model
        self.model.eval()
        self.last_layer_name = last_layer_name
        self.penultimate_layer_name = penultimate_layer_name

        # create feature extractor
        if self.penultimate_layer_name is None:
            self.penultimate_layer_name = list(self.model._modules.keys())[-2]
        self.feature_extractor = create_feature_extractor(self.model, [self.penultimate_layer_name])

        # get the model weights of the last layer
        if last_layer_name is None:
            if hasattr(self.model, "default_cfg"):
                last_layer_name = self.model.default_cfg["classifier"]
            else:
                last_layer_name = list(model._modules.keys())[-1]
            _logger.info("Last layer name: %s", last_layer_name)
        # last_layer = model._modules[last_layer_name]
        last_layer = get_composed_attr(model, last_layer_name.split("."))
        assert isinstance(last_layer, nn.Linear), "Last layer must be a linear layer"

        self.w = last_layer.weight.data.clone()
        self.b = last_layer.bias.data.clone()

        _logger.debug("w shape: %s", self.w.shape)
        _logger.debug("b shape: %s", self.b.shape)

        self.head = list(model._modules.values())[-1]

        # new origin
        self.u = -torch.matmul(torch.linalg.pinv(self.w), self.b).float()
        _logger.debug("New origin shape: %s", self.u.shape)

        self.principal_subspace = None
        self.train_features = []
        self.train_logits = []
        self.alpha = None
        self.top_k = None

    def _get_logits(self, features: Tensor) -> Tensor:
        logits = self.head(features)
        return logits

    def start(self, *args, **kwargs):
        self.principal_subspace = None
        self.train_features = None
        self.train_logits = None
        self.alpha = None
        self.top_k = None

    @torch.no_grad()
    def update(self, x: torch.Tensor, y: torch.Tensor, *args, **kwargs):
        features = self.feature_extractor(x)[self.penultimate_layer_name]
        features = sync_tensor_across_gpus(features)
        if dist.is_initialized():
            dist.gather(features, dst=0)

        if self.train_features is None:
            self.train_features = torch.flatten(features, start_dim=1).cpu()
        else:
            self.train_features = torch.cat([self.train_features, torch.flatten(features, start_dim=1).cpu()], dim=0)

        if self.train_logits is None:
            self.train_logits = self._get_logits(features).cpu()
        else:
            self.train_logits = torch.cat([self.train_logits, self._get_logits(features).cpu()])

    def end(self):
        self.top_k = 1000 if self.train_features.shape[1] > 1500 else 512

        _logger.info("Train features shape: %s", self.train_features.shape)
        _logger.info("Train logits shape: %s", self.train_logits.shape)

        # calculate eigenvectors of the covariance matrix
        ec = EmpiricalCovariance(assume_centered=True)
        ec.fit(self.train_features.cpu().numpy() - self.u.detach().cpu().numpy())
        eig_vals, eigen_vectors = np.linalg.eig(ec.covariance_)
        determinant = np.linalg.det(ec.covariance_)
        _logger.debug("Determinant: %s", determinant)
        _logger.debug("Eigen values: %s", eig_vals)

        # select largest eigenvectors to get the principal subspace
        largest_eigvals_idx = np.argsort(eig_vals * -1)[self.top_k :]
        self.principal_subspace = torch.from_numpy(
            np.ascontiguousarray((eigen_vectors.T[largest_eigvals_idx]).T)
        ).float()
        _logger.debug("Principal subspace: %s", self.principal_subspace)

        # calculate residual
        x_p_t = torch.matmul(self.train_features.cpu() - self.u.cpu(), self.principal_subspace.cpu())
        vlogits = torch.norm(x_p_t, dim=-1)
        self.alpha = self.train_logits.max(dim=-1)[0].mean() / vlogits.mean()
        _logger.debug("Alpha: %s", self.alpha)

        del self.train_features

    @torch.no_grad()
    def __call__(self, x: torch.Tensor) -> torch.Tensor:
        self.feature_extractor = self.feature_extractor.to(x.device)
        features = self.feature_extractor(x)[self.penultimate_layer_name]

        logits = self._get_logits(features)

        x_p_t = torch.norm(
            torch.matmul(torch.flatten(features, 1) - self.u.to(x.device), self.principal_subspace.to(x.device)), dim=-1
        )
        vlogit = x_p_t * self.alpha
        energy = torch.logsumexp(logits, dim=-1)
        score = -vlogit + energy
        return torch.nan_to_num(score, 1e6)



"""
Detection methods.
"""
import logging
import types
from enum import Enum
from functools import partial
from typing import Any, Dict, List

from detectors.methods.templates import Detector, DetectorWrapper

from .csi import CSI
from .dice import Dice
from .doctor import doctor
from .energy import energy
from .entropy import entropy
from .gmm import GMM
from .gradnorm import gradnorm
from .igeood_logits import IgeoodLogits
from .kl_matching import KLMatching
from .knn_cosine import KnnCosine
from .knn_euclides import KnnEuclides
from .knn_projection import KnnProjection
from .mahalanobis import Mahalanobis
from .max_logits import max_logits
from .maxcosine import MaxCosineSimilarity
from .mcdropout import mcdropout
from .msp import msp
from .naive import always_one, always_zero, random_score
from .odin import odin
from .projection import Projection
from .react import ReAct
from .react_projection import ReActProjection
from .relative_mahalanobis import RelativeMahalanobis
from .ssd import SSD
from .vim import ViM

_logger = logging.getLogger(__name__)

detectors_registry = {
    # naive detectors
    "random": random_score,
    "always_one": always_one,
    "always_zero": always_zero,
    # hyperparameter free detectors
    "msp": msp,
    "max_logits": max_logits,
    "kl_matching": KLMatching,
    "vim": ViM,
    "mcdropout": mcdropout,
    "maxcosine": MaxCosineSimilarity,
    "entropy": entropy,
    # hyperparameter detectors
    "odin": odin,
    "doctor": doctor,
    "energy": energy,
    "dice": Dice,
    "react": ReAct,
    "igeood_logits": IgeoodLogits,
    "gradnorm": gradnorm,
    "knn_euclides": KnnEuclides,
    "knn_cosine": KnnCosine,
    "knn_projection": KnnProjection,
    # Features based detectors
    "mahalanobis": Mahalanobis,
    "gmm": GMM,
    "relative_mahalanobis": RelativeMahalanobis,
    "projection": Projection,
    "react_projection": ReActProjection,
    # Special training detectors
    "ssd": SSD,
    "csi": CSI,
    # Not implemented
    "igeood_features": None,
    "bats": None,
    "gram": None,
    "openmax": None,
    "rankfeat": None,
    "godin": None,
}


def register_detector(name: str):
    """Decorator to register a new detector.

    Args:
        name (string): Name of the detector.

    Example::

        @register_detector("my_detector")
        class MyDetector(Detector):
            ...

        detector = create_detector("my_detector")

        @register_detector("my_detector")
        def my_detector(model, **kwargs):
            ...

        detector = create_detector("my_detector")
    """

    def decorator(f):
        detectors_registry[name] = f
        return f

    return decorator


def create_detector(detector_name: str, **kwargs) -> Detector:
    """Create detector factory.

    Args:
        detector_name (string): Name of the detector.
            Already implemented:
                `random`, `msp`, `odin`, `energy`, `mahalanobis`, `react`, `dice`, `knn_euclides`, `igeood_logits`,
                `projection`, `react_projection`, `gradnorm`, `maxcosine`, `mcdropout`, `max_logits`, `kl_matching`,
                `gmm`, `relative_mahalanobis`, `doctor`, `always_one`, `always_zero`, `random_score`, `vim`,
                `entropy`, `ssd`, `csi`, `knn_cosine`, `knn_projection`.

        **kwargs: Additional arguments for the detector.

    Returns:
        Detector: the corresponding detector.
    """
    model = kwargs.pop("model", None)
    if detector_name not in detectors_registry:
        raise ValueError(f"Unknown detector: {detector_name}")
    if not isinstance(detectors_registry[detector_name], types.FunctionType):
        return DetectorWrapper(detectors_registry[detector_name](model=model, **kwargs), **kwargs)
    return DetectorWrapper(partial(detectors_registry[detector_name], model=model, **kwargs), **kwargs)


def list_detectors() -> List[str]:
    """List available detectors.

    Returns:
        List[str]: List of available detectors.
    """
    return list(k for k in detectors_registry.keys() if detectors_registry[k] is not None)


def create_hyperparameters(detector_name: str) -> Dict[str, Any]:
    """Create hyperparameters for the detector.

    Args:
        detector_name (string): Name of the detector.

    Returns:
        Dict[str, Any]: Hyperparameters for the detector.
    """
    import importlib

    try:
        module = importlib.import_module(f"detectors.methods.{detector_name}")
        hyperparameters = module.HYPERPARAMETERS
    except ModuleNotFoundError:
        raise ValueError(f"Unknown detector: {detector_name}")
    except AttributeError:
        hyperparameters = {}
    return hyperparameters


MethodsRegistry = Enum("MethodsRegistry", dict(zip(list_detectors(), list_detectors())))



from setuptools import find_packages, setup

install_requires = [
    "Pillow",
    "numpy",
    "optuna",
    "scikit-learn",
    "scikit-image",
    "timm==0.8.19.dev0",
    "torch>=1.13.1",
    "torchvision",
    "tqdm",
    "scipy",
    "accelerate",
    "psutil",
    "pandas",
    "wilds",
    "faiss-cpu",
    "matplotlib",
]

setup(
    name="detectors",
    version="v0.1.10",
    author="Eduardo Dadalto with the help of all our contributors.",
    author_email="edadaltocg@gmail.com",
    description="Detectors: a python package to benchmark generalized out-of-distribution detection methods.",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="vision deep learning pytorch OOD",
    license="APACHE 2.0",
    url="https://github.com/edadaltocg/detectors",
    package_dir={"": "src"},
    packages=find_packages("src"),
    include_package_data=True,
    python_requires=">=3.8.0",
    install_requires=install_requires,
    extras_require={
        "dev": [
            "black",
            "isort",
            "flake8",
            "pytest",
            "pytest-cov",
            "pytest-timeout",
            "pytest-xdist",
            "twine",
            "sphinx",
            "sphinx_rtd_theme",
            "myst-parser",
            "six",
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
)



import logging
from typing import List, Optional

import faiss
import numpy as np
import torch
from torch import Tensor, nn

from detectors.methods.templates import DetectorWithFeatureExtraction

_logger = logging.getLogger(__name__)

HYPERPARAMETERS = dict(k=dict(low=1, high=100, step=5), avg_topk=[True, False])


class KnnEuclides(DetectorWithFeatureExtraction):
    """K-NN detector based on Euclidean distance.

    Args:
        model (nn.Module): Model to be used to extract features
        features_nodes (Optional[List[str]]): List of strings that represent the feature nodes.
            Defaults to None.
        all_blocks (bool, optional): If True, use all blocks of the model. Defaults to False.
        last_layer (bool, optional): If True, use also the last layer of the model. Defaults to False.
        pooling_op_name (str, optional): Pooling operation to be applied to the features.
            Can be one of ["max", "avg", "flatten", "getitem", "none"]. Defaults to "avg".
        aggregation_method_name (str, optional): Aggregation method to be applied to the features. Defaults to None.
        alpha (float, optional): Alpha parameter for the input pre-processing. Defaults to 1.
        k (int, optional): Number of nearest neighbors to be considered. Defaults to 10.
        avg_topk (bool, optional): If True, average the top-k scores. Defaults to False.

    References:
        [1] https://arxiv.org/abs/2204.06507
    """

    def __init__(
        self,
        model: nn.Module,
        features_nodes: Optional[List[str]] = None,
        all_blocks: bool = False,
        last_layer: bool = False,
        pooling_op_name: str = "avg",
        aggregation_method_name="mean",
        alpha: float = 1,
        k: int = 10,
        avg_topk: bool = False,
        **kwargs,
    ):
        super().__init__(
            model,
            features_nodes=features_nodes,
            all_blocks=all_blocks,
            last_layer=last_layer,
            pooling_op_name=pooling_op_name,
            aggregation_method_name=aggregation_method_name,
        )
        self.alpha = alpha
        self.k = k
        self.mean_op = avg_topk

        assert 0 < self.alpha <= 1, "alpha must be in the interval (0, 1]"

    def _layer_score(self, x: Tensor, layer_name: Optional[str] = None, index: Optional[int] = None):
        x = x / torch.norm(x, p=2, dim=-1, keepdim=True)  # type: ignore
        topk, _ = self.index[layer_name].search(x.cpu().numpy().astype(np.float32), k=self.k)
        topk = torch.from_numpy(topk).to(x.device)
        # pairwise = torch.cdist(x, self.ref[layer_name].to(x.device), p=2)
        # topk, _ = torch.topk(pairwise, k=self.k, dim=-1, largest=False)
        if self.mean_op:
            return -topk.mean(dim=-1)
        else:
            return -topk[:, -1]

    def _fit_params(self) -> None:
        self.ref = {}
        self.index = {}
        for layer_name, layer_features in self.train_features.items():
            self.ref[layer_name] = layer_features[
                torch.randperm(layer_features.shape[0])[: int(self.alpha * layer_features.shape[0])]
            ]
            # normalize train features
            self.ref[layer_name] = self.ref[layer_name] / torch.norm(
                self.ref[layer_name], p="fro", dim=-1, keepdim=True
            )
            self.index[layer_name] = faiss.IndexFlatL2(self.ref[layer_name].shape[1])
            self.index[layer_name].add(self.ref[layer_name].cpu().numpy().astype(np.float32))



import logging

import numpy as np
import sklearn.metrics as skm
from sklearn.linear_model import LogisticRegressionCV

_logger = logging.getLogger(__name__)


class WeightRegression:
    def __init__(
        self,
        verbose=False,
        ignore_dim=0,
        split_size=1000,
        regressor=LogisticRegressionCV,
        recall_level=0.95,
        seed=0,
    ):
        np.random.seed(seed)
        self.verbose = verbose
        self.split_size = split_size
        self.regressor = regressor(
            n_jobs=-1
            # scoring=skm.make_scorer(self.scoring_obj, greater_is_better=False)
        )
        self.__name__ = "weight_regression"
        self.ignore_dim = ignore_dim
        self.recall_level = recall_level

    def __call__(self, in_data: np.ndarray, out_data: np.ndarray, *args, **kwargs):
        in_data = in_data[:, self.ignore_dim :]
        out_data = out_data[:, self.ignore_dim :]

        # split
        x_train, y_train, x_test, y_test = self.split_data(in_data, out_data)
        # fit
        self.regressor.fit(x_train, y_train)
        _logger.info(f"Coeficients: {self.regressor.coef_}")

        # predict
        y_pred_test = self.regressor.predict_proba(x_test)[:, 1]
        # evaluate
        if self.verbose:
            print(self.regressor.coef_)
            y_pred_train = self.regressor.predict_proba(x_train)[:, 1]
            _logger.info(
                "training fpr: {:.4f}".format(self.scoring_obj(y_train, y_pred_train)),
            )
            _logger.info(
                "test fpr: {:.4f}".format(self.scoring_obj(y_test, y_pred_test)),
            )
        in_scores = y_pred_test[y_test == 1]
        out_scores = y_pred_test[y_test == 0]

        return in_scores, out_scores

    def split_data(self, data_in, data_out):
        mask_train = slice(0, self.split_size)
        mask_test = slice(self.split_size, 100000)
        # Train
        if len(data_in.shape) == 1:
            data_in = data_in.reshape(-1, 1)
            data_out = data_out.reshape(-1, 1)

        x1_train = data_in[mask_train]
        x2_train = data_out[mask_train]
        x_train = np.concatenate((x1_train, x2_train))
        y_train = np.concatenate((np.ones((len(x1_train), 1)), np.zeros((len(x2_train), 1)))).reshape(-1)

        # Test
        x1_test = data_in[mask_test]
        x2_test = data_out[mask_test]
        x_test = np.concatenate((x1_test, x2_test))
        y_test = np.concatenate((np.ones((len(x1_test), 1)), np.zeros((len(x2_test), 1)))).reshape(-1)
        return x_train, y_train, x_test, y_test

    def scoring_obj(self, y_true, y_pred):
        fprs, tprs, thresholds = skm.roc_curve(y_true, y_pred, pos_label=1)



# TODO
import torch


def iterate_data_rankfeat(data_loader, model, temperature):
    confs = []
    for b, (x, y) in enumerate(data_loader):
        if b % 100 == 0:
            print("{} batches processed".format(b))
        inputs = x.cuda()

        # Logit of Block 4 feature
        feat1 = model.module.intermediate_forward(inputs, layer_index=4)
        B, C, H, W = feat1.size()
        feat1 = feat1.view(B, C, H * W)
        u, s, v = torch.linalg.svd(feat1, full_matrices=False)
        feat1 = feat1 - s[:, 0:1].unsqueeze(2) * u[:, :, 0:1].bmm(v[:, 0:1, :])
        # if you want to use PI for acceleration, comment the above 2 lines and uncomment the line below
        # feat1 = feat1 - power_iteration(feat1, iter=20)
        feat1 = feat1.view(B, C, H, W)
        logits1 = model.module.head(model.module.before_head(feat1))

        # Logit of Block 4 feature
        feat2 = model.module.intermediate_forward(inputs, layer_index=3)
        B, C, H, W = feat2.size()
        feat2 = feat2.view(B, C, H * W)
        u, s, v = torch.linalg.svd(feat2, full_matrices=False)
        feat2 = feat2 - s[:, 0:1].unsqueeze(2) * u[:, :, 0:1].bmm(v[:, 0:1, :])
        # if you want to use PI for acceleration, comment the above 2 lines and uncomment the line below
        # feat2 = feat2 - power_iteration(feat2, iter=20)
        feat2 = feat2.view(B, C, H, W)
        feat2 = model.module.body.block4(feat2)
        logits2 = model.module.head(model.module.before_head(feat2))

        # Fusion at the logit space
        logits = (logits1 + logits2) / 2
        conf = temperature * torch.logsumexp(logits / temperature, dim=1)
        confs.extend(conf.data.cpu().numpy())

    return confs



import json
import logging
import os
from typing import Any, Dict, List

import pandas as pd
import torch
import torch.distributed as dist
from torch.utils.data import Dataset

_logger = logging.getLogger(__name__)


def str_to_dict(string: str) -> Dict[str, Any]:
    string = string.strip().strip("'").replace("'", '"')
    return json.loads(string)


def append_results_to_csv_file(results, filename="results.csv"):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    results = {k: [v] for k, v in results.items()}
    results = pd.DataFrame.from_dict(results, orient="columns")

    if not os.path.isfile(filename):
        results.to_csv(filename, header=True, index=False)
    else:  # it exists, so append without writing the header
        results.to_csv(filename, mode="a", header=False, index=False)


class ConcatDatasetsDim1(Dataset):
    def __init__(self, datasets: List[Dataset]):
        self.datasets = datasets

        assert all(len(ds) for ds in self.datasets)

    def __getitem__(self, index):
        outputs = []
        for dataset in self.datasets:
            elem = dataset.__getitem__(index)
            if isinstance(elem, (list, tuple)):
                outputs += elem
            else:
                outputs += [elem]
        return tuple(outputs)

    def __len__(self):
        return len(self.datasets[0])


def sync_tensor_across_gpus(t: torch.Tensor) -> torch.Tensor:
    """Gather tensor from all gpus and return a tensor with dim 0 equal to the number of gpus.

    Args:
        t (torch.Tensor): _description_

    Returns:
        torch.Tensor: _description_

    References:
        https://discuss.pytorch.org/t/ddp-evaluation-gather-output-loss-and-stuff-how-to/130593/2
    """
    if not dist.is_initialized():
        return t
    group = dist.group.WORLD
    group_size = dist.get_world_size(group)
    if group_size == 1:
        return t
    gather_t_tensor = [torch.zeros_like(t) for _ in range(group_size)]
    dist.all_gather(gather_t_tensor, t)  # this works with nccl backend when tensors need to be on gpu.
    # for gloo and mpi backends, tensors need to be on cpu. also this works single machine with
    # multiple   gpus. for multiple nodes, you should use dist.all_gather_multigpu. both have the
    # same definition... see [here](https://pytorch.org/docs/stable/distributed.html).
    #  somewhere in the same page, it was mentioned that dist.all_gather_multigpu is more for
    # multi-nodes. still dont see the benefit of all_gather_multigpu. the provided working case in
    # the doc is  vague...
    # move tensors to cpu
    # gather_t_tensor = [t.cpu() for t in gather_t_tensor]
    gather_t_tensor = torch.cat(gather_t_tensor, dim=0)

    return gather_t_tensor



import logging

import torch
from torch import Tensor

_logger = logging.getLogger(__name__)


class CosineAggregation:
    def __init__(self, *args, **kwargs) -> None:
        self.mu = None

    def fit(self, stack: Tensor, *args, **kwargs):
        self.mu = stack.mean(dim=0, keepdim=True)

    def __call__(self, scores: Tensor, *args, **kwargs):
        return torch.nn.functional.cosine_similarity(scores, self.mu.to(scores.device), dim=1)



import torch
import torch.nn as nn


class SupConLoss(nn.Module):
    """Supervised Contrastive Learning.
    It also supports the unsupervised contrastive loss in
    SimCLR: A Simple Framework for Contrastive Learning of Visual Representations

    References:
        [1] https://github.com/HobbitLong/SupContrast
        [2] https://arxiv.org/pdf/2004.11362.pdf
        [3] https://arxiv.org/abs/2002.05709
    """

    def __init__(self, temperature=0.07, contrast_mode="all", base_temperature=0.07):
        super(SupConLoss, self).__init__()
        self.temperature = temperature
        self.contrast_mode = contrast_mode
        self.base_temperature = base_temperature

    def forward(self, features, labels=None, mask=None):
        """Compute loss for model. If both `labels` and `mask` are None,
        it degenerates to SimCLR unsupervised loss.

        Args:
            features: hidden vector of shape [bsz, n_views, ...].
            labels: ground truth of shape [bsz].
            mask: contrastive mask of shape [bsz, bsz], mask_{i,j}=1 if sample j
                has the same class as sample i. Can be asymmetric.
        Returns:
            A loss scalar.
        """
        device = torch.device("cuda") if features.is_cuda else torch.device("cpu")

        if len(features.shape) < 3:
            raise ValueError("`features` needs to be [bsz, n_views, ...]," "at least 3 dimensions are required")
        if len(features.shape) > 3:
            features = features.view(features.shape[0], features.shape[1], -1)

        batch_size = features.shape[0]
        if labels is not None and mask is not None:
            raise ValueError("Cannot define both `labels` and `mask`")
        elif labels is None and mask is None:
            mask = torch.eye(batch_size, dtype=torch.float32).to(device)
        elif labels is not None:
            labels = labels.contiguous().view(-1, 1)
            if labels.shape[0] != batch_size:
                raise ValueError("Num of labels does not match num of features")
            mask = torch.eq(labels, labels.T).float().to(device)
        else:
            mask = mask.float().to(device)

        contrast_count = features.shape[1]
        contrast_feature = torch.cat(torch.unbind(features, dim=1), dim=0)
        if self.contrast_mode == "one":
            anchor_feature = features[:, 0]
            anchor_count = 1
        elif self.contrast_mode == "all":
            anchor_feature = contrast_feature
            anchor_count = contrast_count
        else:
            raise ValueError("Unknown mode: {}".format(self.contrast_mode))

        # compute logits
        anchor_dot_contrast = torch.div(torch.matmul(anchor_feature, contrast_feature.T), self.temperature)
        # for numerical stability
        logits_max, _ = torch.max(anchor_dot_contrast, dim=1, keepdim=True)
        logits = anchor_dot_contrast - logits_max.detach()

        # tile mask
        mask = mask.repeat(anchor_count, contrast_count)
        # mask-out self-contrast cases
        logits_mask = torch.scatter(
            torch.ones_like(mask), 1, torch.arange(batch_size * anchor_count).view(-1, 1).to(device), 0
        )
        mask = mask * logits_mask

        # compute log_prob
        exp_logits = torch.exp(logits) * logits_mask
        log_prob = logits - torch.log(exp_logits.sum(1, keepdim=True))

        # compute mean of log-likelihood over positive
        mean_log_prob_pos = (mask * log_prob).sum(1) / mask.sum(1)

        # loss
        loss = -(self.temperature / self.base_temperature) * mean_log_prob_pos
        loss = loss.view(anchor_count, batch_size).mean()

        return loss


class CSILoss(SupConLoss):
    """Contrasting shifted instances (con-SI) loss.

    References:
        [1] https://arxiv.org/abs/2007.08176
    """

    def __init__(self, temperature=0.07, contrast_mode="all", base_temperature=0.07, lbd=1):
        super().__init__(temperature=temperature, contrast_mode=contrast_mode, base_temperature=base_temperature)

        self.lbd = lbd
        self.shift_criterion = nn.CrossEntropyLoss()

    def forward(self, features, shift_labels, labels=None, mask=None):
        loss_sim = super().forward(features, labels, mask)
        loss_shift = self.shift_criterion(features, shift_labels)
        return loss_sim + self.lbd * loss_shift
